# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import netaddr

from neutron.ipam import utils as ipam_utils
from neutron.tests.unit import testlib_api
from neutron_lib import context

from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.common import ipam
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi
from networking_infoblox.tests import base


class IpamControllerTestHelper(object):

    DEFAULT_OPTIONS = {'network_name': 'test-network',
                       'external': False,
                       'shared': False,
                       'subnet_name': 'test-subnet',
                       'cidr': '11.11.1.0/24',
                       'dns_nameservers': [],
                       'enable_dhcp': True,
                       'network_view': 'test-network-view',
                       'network_view_exists': False,
                       'network_exists': False}

    def __init__(self):
        self.neutron_cxt = context.get_admin_context()
        self.tenant_id = self.neutron_cxt.tenant_id
        self.network = None
        self.subnet = None
        self.ib_cxt = mock.Mock()
        self.ib_cxt.tenant_id = self.tenant_id
        self.ib_cxt.tenant_name = 'test_tenant'
        self.ib_cxt.context = self.neutron_cxt
        self.ib_cxt.grid_config.admin_network_deletion = False
        self.ib_cxt.grid_config.network_template = None
        self.ib_cxt.grid_config.dhcp_relay_management_network = None
        self.ib_cxt.grid_config.wapi_version = '2.3'
        self.ib_cxt.grid_config.default_domain_name_pattern = 'int.test.com'
        self.ib_cxt.grid_config.external_domain_name_pattern = 'ext.test.com'
        self.ib_cxt.grid_config.zone_creation_strategy = []

    def prepare_test(self, options):
        self.options = self._get_options(options)

        self.network = self.create_network(self.options['network_name'],
                                           self.options['external'],
                                           self.options['shared'])

        self.subnet = self.create_subnet(self.options['subnet_name'],
                                         self.network['id'],
                                         self.options['cidr'],
                                         self.options['dns_nameservers'],
                                         self.options['enable_dhcp'])

        self.update_context(self.network,
                            self.subnet,
                            self.options['network_view'],
                            self.options['network_view_exists'])

    def _get_options(self, options):
        if not isinstance(options, dict):
            raise ValueError('Options should be passed as dict')
        for key in self.DEFAULT_OPTIONS:
            if key not in options:
                options[key] = self.DEFAULT_OPTIONS[key]
        return options

    def create_network(self, name, external=False, shared=False):
        return {
            'status': 'ACTIVE',
            'subnets': [],
            'name': name,
            'provider:physical_network': 'None',
            'router:external': external,
            'tenant_id': self.tenant_id,
            'admin_state_up': True,
            'mtu': 0,
            'shared': shared,
            'provider:network_type': 'vxlan',
            'id': str.format("{}-id", name),
            'provider:segmentation_id': 1001
        }

    def create_subnet(self, name, network_id, cidr, dns_nameservers,
                      enable_dhcp=True):
        ip_network = netaddr.IPNetwork(cidr)
        ip_range = list(ip_network)
        if ip_network.version == 6:
            raise ValueError("Test with IPv4 CIDR only")

        return {
            'name': 'subn01',
            'enable_dhcp': enable_dhcp,
            'network_id': network_id,
            'tenant_id': self.tenant_id,
            'dns_nameservers': dns_nameservers,
            'ipv6_ra_mode': 'None',
            'allocation_pools': ipam_utils.generate_pools(cidr, ip_range[1]),
            'gateway_ip': ip_range[1],
            'ipv6_address_mode': 'None',
            'ip_version': ip_network.version,
            'host_routes': [],
            'cidr': cidr,
            'id': str.format("{}-id", name),
            'subnetpool_id': 'None'
        }

    def update_context(self, network, subnet, network_view,
                       network_view_exists=False):
        self.ib_cxt.network = network
        self.ib_cxt.subnet = subnet
        self.ib_cxt.mapping.network_view = network_view
        self.ib_cxt.mapping.dns_view = 'test-view'
        if network_view_exists is False:
            self.ib_cxt.mapping.network_view_id = None
            self.ib_cxt.mapping.authority_member = None
            self.ib_cxt.mapping.dhcp_members = []
            self.ib_cxt.mapping.dns_members = []
        else:
            self.ib_cxt.mapping.network_view_id = str.format(
                "networkview/ZG5zLm5ldHdvcmtfdmlldyQw:{}/false",
                network_view)
            self.ib_cxt.mapping.authority_member = mock.Mock()
            self.ib_cxt.mapping.dhcp_members = [mock.Mock()]
            self.ib_cxt.mapping.dns_members = [mock.Mock()]


class IpamSyncControllerTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):
        super(IpamSyncControllerTestCase, self).setUp()
        self.helper = IpamControllerTestHelper()
        self.ib_cxt = self.helper.ib_cxt
        self.grid_config = self.ib_cxt.grid_config

    def validate_network_creation(self, network_view, subnet):
        if self.ib_cxt.mapping.network_view_id:
            self.ib_cxt.ibom.create_network_view.assert_not_called()
        else:
            self.ib_cxt.reserve_authority_member.assert_called_once_with()

        self.ib_cxt.reserve_service_members.assert_called_once_with()

        self.ib_cxt.ibom.create_network.assert_called_once_with(
            network_view, subnet['cidr'], mock.ANY, mock.ANY,
            subnet['gateway_ip'], None, mock.ANY)

        allocation_pools = subnet['allocation_pools'][0]
        first_ip = netaddr.IPAddress(allocation_pools.first,
                                     subnet['ip_version']).format()
        last_ip = netaddr.IPAddress(allocation_pools.last,
                                    subnet['ip_version']).format()
        self.ib_cxt.ibom.create_ip_range.assert_called_once_with(
            network_view,
            first_ip,
            last_ip,
            subnet['cidr'],
            True,
            mock.ANY)

    @mock.patch.object(ib_objects, 'NetworkView')
    def test_create_subnet_new_network_view_nwview_preexist(self, mck_nw_view):
        mck_nw_view.create_check_exists.return_value = mock.Mock(), False
        self._test_create_subnet_new_network_view()

    @mock.patch.object(ib_objects, 'NetworkView')
    def test_create_subnet_new_network_view_nwview_created(self, mck_nw_view):
        mck_nw_view.create_check_exists.return_value = mock.Mock(), True
        self._test_create_subnet_new_network_view()

    @mock.patch.object(dbi, 'update_network_view_id', mock.Mock())
    @mock.patch.object(dbi, 'associate_network_view', mock.Mock())
    @mock.patch('infoblox_client.objects.Tenant')
    @mock.patch('infoblox_client.objects.IPRange')
    def _test_create_subnet_new_network_view(self, ip_range_mock, tenant_mock):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ip_range_mock.search = mock.Mock(return_value=None)
        tenant = mock.Mock()
        tenant_mock.search = mock.Mock(return_value=tenant)
        with mock.patch.object(ib_objects.Network,
                               'search',
                               return_value=None):
            ipam_controller._register_mapping_member = mock.Mock()
            rollback_list = []
            ipam_controller.create_subnet(rollback_list)

        tenant_mock.search.assert_called_once_with(
            self.ib_cxt.connector, id=self.ib_cxt.tenant_id)
        tenant.update.assert_called_once_with()
        assert tenant.name == self.ib_cxt.tenant_name

        self.validate_network_creation(self.helper.options['network_view'],
                                       self.helper.subnet)

    @mock.patch.object(dbi, 'update_network_view_id', mock.Mock())
    @mock.patch.object(dbi, 'associate_network_view', mock.Mock())
    @mock.patch('infoblox_client.objects.Tenant')
    @mock.patch('infoblox_client.objects.Member')
    @mock.patch('infoblox_client.objects.IPRange')
    def test_create_subnet_existing_network_view(self, ip_range_mock,
                                                 member_mock, tenant_mock):
        test_opts = {'cidr': '12.12.12.0/24', 'network_view_exists': True}
        self.helper.prepare_test(test_opts)

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        member_mock.search = mock.Mock(return_value=None)
        ip_range_mock.search = mock.Mock(return_value=None)
        tenant = mock.Mock()
        tenant_mock.search = mock.Mock(return_value=tenant)
        with mock.patch.object(ib_objects.Network,
                               'search',
                               return_value=None):
            ipam_controller._register_mapping_member = mock.Mock()
            rollback_list = []
            ipam_controller.create_subnet(rollback_list)

        tenant_mock.search.assert_called_once_with(
            self.ib_cxt.connector, id=self.ib_cxt.tenant_id)
        tenant.update.assert_called_once_with()
        assert tenant.name == self.ib_cxt.tenant_name

        self.validate_network_creation(self.helper.options['network_view'],
                                       self.helper.subnet)

    @mock.patch.object(ib_objects, 'NetworkView')
    def test_create_subnet_existing_private_network_nwview_preexist(
            self, mck_nw_view):
        mck_nw_view.create_check_exists.return_value = mock.Mock(), False
        self._test_create_subnet_existing_private_network()

    @mock.patch.object(ib_objects, 'NetworkView')
    def test_create_subnet_existing_private_network_nwview_created(
            self, mck_nw_view):
        mck_nw_view.create_check_exists.return_value = mock.Mock(), True
        self._test_create_subnet_existing_private_network()

    @mock.patch.object(dbi, 'update_network_view_id', mock.Mock())
    @mock.patch.object(dbi, 'associate_network_view', mock.Mock())
    def _test_create_subnet_existing_private_network(self):
        test_opts = {'network_exists': True,
                     'external': False,
                     'shared': False}
        self.helper.prepare_test(test_opts)

        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False
        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        with mock.patch.object(ib_objects.Network,
                               'search',
                               return_value=mock.Mock()):
            self.assertRaises(exc.InfobloxPrivateSubnetAlreadyExist,
                              ipam_controller.create_subnet,
                              [])

    @mock.patch.object(ib_objects, 'NetworkView')
    def test_create_subnet_existing_external_network_nwview_preexist(
            self, mck_nw_view):
        mck_nw_view.create_check_exists.return_value = mock.Mock(), False
        self._test_create_subnet_existing_external_network()

    @mock.patch.object(ib_objects, 'NetworkView')
    def test_create_subnet_existing_external_network_nwview_created(
            self, mck_nw_view):
        mck_nw_view.create_check_exists.return_value = mock.Mock(), True
        self._test_create_subnet_existing_external_network()

    @mock.patch.object(dbi, 'update_network_view_id', mock.Mock())
    @mock.patch.object(dbi, 'associate_network_view', mock.Mock())
    @mock.patch('infoblox_client.objects.Tenant')
    def _test_create_subnet_existing_external_network(self, tenant_mock):
        test_opts = {'network_name': 'extnet',
                     'subnet_name': 'extsub',
                     'cidr': '172.192.1.0/24',
                     'external': True,
                     'network_exists': True}
        self.helper.prepare_test(test_opts)

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller._allocate_pools = mock.Mock()
        tenant = mock.Mock()
        tenant_mock.search = mock.Mock(return_value=tenant)
        with mock.patch.object(ib_objects.Network,
                               'search',
                               return_value=mock.Mock()):
            rollback_list = []
            ipam_controller.create_subnet(rollback_list)

        tenant_mock.search.assert_called_once_with(
            self.ib_cxt.connector, id=self.ib_cxt.tenant_id)
        tenant.update.assert_called_once_with()
        assert tenant.name == self.ib_cxt.tenant_name

        self.ib_cxt.ibom.update_network_options.assert_called_once_with(
            mock.ANY, mock.ANY)

    def test_update_subnet_allocation_pools(self):
        test_opts = {'network_exists': True,
                     'external': False,
                     'shared': False}
        self.helper.prepare_test(test_opts)
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False

        new_pools = (netaddr.IPRange('11.11.1.25', '11.11.1.30'),
                     netaddr.IPRange('11.11.1.45', '11.11.1.60'))
        self.ib_cxt.subnet['allocation_pools'] = new_pools
        ip_version = self.ib_cxt.subnet['ip_version']

        connector = mock.Mock()
        ib_pools = (ib_objects.IPRange(connector,
                                       start_addr='11.11.1.3',
                                       end_addr='11.11.1.19'),
                    ib_objects.IPRange(connector,
                                       start_addr='11.11.1.25',
                                       end_addr='11.11.1.30'))

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        rollback_list = []
        with mock.patch.object(ib_objects.IPRange,
                               'search_all',
                               return_value=ib_pools):
            ipam_controller.update_subnet_allocation_pools(rollback_list)

            # 1st range from ib_pools should be removed
            ib_pools[0].connector.delete_object.assert_called_once_with(None)

            # 2nd pool from new_pools should be added
            self.ib_cxt.ibom.create_ip_range.assert_called_once_with(
                self.helper.options['network_view'],
                netaddr.IPAddress(new_pools[1].first, ip_version).format(),
                netaddr.IPAddress(new_pools[1].last, ip_version).format(),
                self.helper.subnet['cidr'],
                mock.ANY, mock.ANY)

    def test_delete_subnet_for_private_network(self):
        test_opts = {'network_exists': True,
                     'external': False,
                     'shared': False}
        self.helper.prepare_test(test_opts)

        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False
        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller._release_service_members = mock.Mock()
        with mock.patch.object(ib_objects.Network,
                               'search_all',
                               return_value=[]):
            ipam_controller.delete_subnet()

        self.ib_cxt.ibom.delete_network.assert_called_once_with(
            self.helper.options['network_view'], self.helper.subnet['cidr'])

    def _create_ib_network_ea(self):
        network_ea = {'CMP Type': {'value': 'OpenStack'},
                      'Cloud API Owned': {'value': 'True'},
                      'Tenant ID': {'value': 'test-id'},
                      'Tenant Name': {'value': 'tenant-name'},
                      'Account': {'value': 'admin'},
                      'Is External': {'value': 'False'},
                      'Is Shared': {'value': 'True'},
                      'Network ID': {'value': 'True'},
                      'Network Name': {'value': 'True'},
                      'Subnet ID': {'value': 'True'},
                      'Subnet Name': {'value': 'True'},
                      'Network Encap': {'value': 'gre'},
                      'Segmentation ID': {'value': 'segmentation-id'},
                      'Physical Network Name': {'value': 'physical-network'}}
        return ib_objects.EA.from_dict(network_ea)

    def _create_ib_range_ea(self):
        range_ea = {'CMP Type': {'value': 'OpenStack'},
                    'Cloud API Owned': {'value': 'True'},
                    'Tenant ID': {'value': 'test-id'},
                    'Tenant Name': {'value': 'tenant-name'},
                    'Account': {'value': 'admin'}}
        return ib_objects.EA.from_dict(range_ea)

    def _reset_ib_range_ea(self):
        expected_ea = {'Cloud API Owned': {'value': 'False'},
                       'CMP Type': {'value': 'N/A'},
                       'Tenant ID': {'value': 'N/A'}}
        return expected_ea

    def test_delete_subnet_for_external_network_not_deletable(self):
        test_opts = {'external': True, 'network_exists': True}
        self.helper.prepare_test(test_opts)
        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller._release_service_members = mock.Mock()

        ib_network_ea = self._create_ib_network_ea()
        ib_network_mock = mock.Mock(extattrs=ib_network_ea)
        ib_range_ea = self._create_ib_range_ea()
        ib_ranges_mock = [mock.Mock(extattrs=ib_range_ea)]
        expected_ea = self._reset_ib_range_ea()

        with mock.patch.object(ib_objects.Network,
                               'search_all',
                               return_value=[]):
            with mock.patch.object(ib_objects.IPRange,
                                   'search_all',
                                   return_value=ib_ranges_mock):
                ipam_controller.delete_subnet(ib_network_mock)
                assert ib_network_mock.update.called
                assert ib_network_mock.extattrs.to_dict() == expected_ea
                assert ib_ranges_mock[0].update.called
                assert ib_ranges_mock[0].extattrs.to_dict() == expected_ea

    def test_delete_subnet_for_external_network_deletable(self):
        test_opts = {'external': True, 'network_exists': True}
        self.helper.prepare_test(test_opts)
        self.grid_config.admin_network_deletion = True

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller._release_service_members = mock.Mock()
        with mock.patch.object(ib_objects.Network,
                               'search_all',
                               return_value=[]):
            ipam_controller.delete_subnet()

        self.ib_cxt.ibom.delete_network.assert_called_once_with(
            self.helper.options['network_view'], self.helper.subnet['cidr'])

    def test_allocate_specific_ip(self):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        ip_address = '11.11.1.3'
        mac = ':'.join(['00'] * 6)
        dns_view = self.ib_cxt.mapping.dns_view
        zone_auth = 'ib.com'
        hostname = mock.ANY
        ea_ip_address = mock.ANY

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller.pattern_builder = mock.Mock()
        ipam_controller.pattern_builder.get_zone_name.return_value = zone_auth

        ipam_controller.allocate_specific_ip(ip_address, mac)

        ipam_controller.pattern_builder.get_zone_name.assert_called_once_with(
            is_external=self.ib_cxt.network_is_external)

        self.ib_cxt.ip_alloc.allocate_given_ip.assert_called_once_with(
            self.helper.options['network_view'], dns_view, zone_auth,
            hostname, mac, ip_address, ea_ip_address)

    def test_allocate_ip_from_pool(self):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        subnet_id = 'subnet-id'
        allocation_pools = [
            {'start': '11.11.1.1', 'end': '11.11.1.150'},
            {'start': '11.11.1.151', 'end': '11.11.1.253'}]
        mac = ':'.join(['00'] * 6)
        dns_view = self.ib_cxt.mapping.dns_view
        zone_auth = 'ib.com'
        hostname = mock.ANY
        ea_ip_address = mock.ANY

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller.pattern_builder = mock.Mock()
        ipam_controller.pattern_builder.get_zone_name.return_value = zone_auth

        ipam_controller.allocate_ip_from_pool(subnet_id, allocation_pools, mac)

        ipam_controller.pattern_builder.get_zone_name.assert_called_once_with(
            is_external=self.ib_cxt.network_is_external)

        self.ib_cxt.ip_alloc.allocate_ip_from_range.assert_called_once_with(
            self.helper.options['network_view'], dns_view, zone_auth,
            hostname, mac, allocation_pools[0]['start'],
            allocation_pools[0]['end'], ea_ip_address)

    def test_deallocate_ip(self):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        ip_address = '11.11.1.1'
        dns_view = self.ib_cxt.mapping.dns_view

        ipam_controller = ipam.IpamSyncController(self.ib_cxt)
        ipam_controller.deallocate_ip(ip_address)

        self.ib_cxt.ip_alloc.deallocate_ip.assert_called_once_with(
            self.helper.options['network_view'], dns_view, ip_address)


class IpamAsyncControllerTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):
        super(IpamAsyncControllerTestCase, self).setUp()
        self.helper = IpamControllerTestHelper()
        self.ib_cxt = self.helper.ib_cxt

    def test_update_network_sync_without_subnet(self):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        ipam_controller = ipam.IpamAsyncController(self.ib_cxt)
        with mock.patch.object(dbi,
                               'get_subnets_by_network_id',
                               return_value=[]):
            ipam_controller.update_network_sync()
            assert not self.ib_cxt.ibom.get_network.called
            assert not self.ib_cxt.ibom.update_network_options.called

    @mock.patch.object(dbi, 'get_network_view_mappings', return_value=[])
    def test_update_network_sync_without_network_view_mapping(self, dbi_mock):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        ipam_controller = ipam.IpamAsyncController(self.ib_cxt)
        with mock.patch.object(dbi,
                               'get_subnets_by_network_id',
                               return_value=[{'id': 'subnet-id',
                                              'cidr': '11.11.1.0/24',
                                              'network_id': 'test_net'}]):
            ipam_controller.update_network_sync()
            assert not self.ib_cxt.ibom.get_network.called
            assert not self.ib_cxt.ibom.update_network_options.called

    @mock.patch.object(utils, 'find_one_in_list')
    @mock.patch.object(dbi, 'get_network_view_mappings')
    def test_update_network_sync_with_network_view_mapping(
            self, netview_mapping_mock, find_row_mock):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        test_netview_mapping = utils.json_to_obj(
            'NetworkViewMapping',
            {'network_view_id': 'test-id', 'network_view': 'test-view'})
        test_netview = utils.json_to_obj(
            'NetworkView',
            {'id': 'test-id', 'network_view': 'test-view'})
        netview_mapping_mock.return_value = [test_netview_mapping]
        find_row_mock.return_value = test_netview

        ipam_controller = ipam.IpamAsyncController(self.ib_cxt)
        with mock.patch.object(dbi,
                               'get_subnets_by_network_id',
                               return_value=[{'id': 'subnet-id',
                                              'cidr': '11.11.1.0/24',
                                              'network_id': 'test_net'}]):
            ipam_controller.update_network_sync()
            assert self.ib_cxt.ibom.get_network.called
            assert self.ib_cxt.ibom.update_network_options.called

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch.object(dbi, 'get_network_view_mappings', return_value=[])
    def test_update_network_update_zone(self, dbi_mock, dns_mock):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        dns_controller = mock.Mock()
        dns_mock.return_value = dns_controller
        ipam_controller = ipam.IpamAsyncController(self.ib_cxt)
        with mock.patch.object(dbi,
                               'get_subnets_by_network_id',
                               return_value=[{'id': 'subnet-id',
                                              'cidr': '11.11.1.0/24',
                                              'network_id': 'test_net'}]):
            ipam_controller.update_network_sync()
            dns_controller.update_dns_zones.assert_called_once_with()

    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch.object(dbi, 'get_network_view_mappings', return_value=[])
    def test_update_network_update_zone_names(self, dbi_mock, dns_mock,
                                              cxt_mock):
        test_opts = dict()
        self.helper.prepare_test(test_opts)

        dns_controller = mock.Mock()
        dns_mock.return_value = dns_controller
        ipam_controller = ipam.IpamAsyncController(self.ib_cxt)
        cxt_mock.return_value = self.ib_cxt
        with mock.patch.object(dbi,
                               'get_subnets_by_network_id',
                               return_value=[{'id': 'subnet-id',
                                              'cidr': '11.11.1.0/24',
                                              'network_id': 'test_net'}]):
            ipam_controller.update_network_sync(need_new_zones=True)
            dns_controller.update_dns_zones.assert_not_called()
            dns_controller.create_dns_zones.assert_called_once_with([])


class IpamSyncControllerUnitTestCase(base.TestCase):

    def setUp(self):
        super(IpamSyncControllerUnitTestCase, self).setUp()

    def create_restart_data(self, allow_service_restart=True):
        helper = IpamControllerTestHelper()
        ib_cxt = helper.ib_cxt
        ib_cxt.grid_config.allow_service_restart = allow_service_restart
        ib_cxt.grid_config.dhcp_support = True
        ib_cxt.ibom.restart_all_services = mock.Mock()
        ipam_controller = ipam.IpamSyncController(ib_cxt)
        member = {'name': 'member1'}
        mock.patch.object(ipam_controller, '_get_service_members',
                          return_value=[member['name']]).start()
        return ib_cxt, ipam_controller, member

    def test_restart_services_positive(self):
        (ib_cxt, ipam_controller, member) = self.create_restart_data()
        with mock.patch.object(ib_objects.Member, 'search',
                               return_value=member):
            ipam_controller._restart_services()
            ib_objects.Member.search.assert_called_once_with(
                ib_cxt.connector, host_name=member['name'],
                return_fields=['host_name'])
        ipam_controller._get_service_members.assert_called_once_with(
            'member_name')
        ib_cxt.ibom.restart_all_services.assert_called_once_with(member)

    def test_restart_services_negative(self):
        (ib_cxt, ipam_controller, member) = self.create_restart_data(False)
        with mock.patch.object(ib_objects.Member, 'search',
                               return_value=member):
            ipam_controller._restart_services()
            ib_objects.Member.search.assert_not_called()
        ipam_controller._get_service_members.assert_not_called()
        ib_cxt.ibom.restart_all_services.assert_not_called()



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_serialization import jsonutils

from neutron.tests.unit import testlib_api
from neutron_lib import context

from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import member
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi

from networking_infoblox.tests import base


class GridMemberTestCase(base.TestCase, testlib_api.SqlTestCase):
    def setUp(self):

        super(GridMemberTestCase, self).setUp()
        self.ctx = context.get_admin_context()

        self.test_grid_config = grid.GridConfiguration(self.ctx)
        self.test_grid_config.gm_connector = mock.Mock()
        self.test_grid_config.grid_id = 100
        self.test_grid_config.grid_name = "Test Grid 1"
        self.test_grid_config.grid_master_host = '192.168.1.7'
        self.test_grid_config.grid_master_name = 'nios-7.2.0-master.com'
        self.test_grid_config.admin_user_name = 'admin'
        self.test_grid_config.admin_password = 'infoblox'
        self.test_grid_config.wapi_version = '2.3'

    def test_sync_grid(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        grids = dbi.get_grids(self.ctx.session)
        self.assertEqual(1, len(grids))
        self.assertEqual(self.test_grid_config.grid_id, grids[0]['grid_id'])
        self.assertEqual(self.test_grid_config.grid_name,
                         grids[0]['grid_name'])
        expected_grid_connection = {
            "wapi_version": self.test_grid_config.wapi_version,
            "ssl_verify": self.test_grid_config.ssl_verify,
            "http_pool_connections":
                self.test_grid_config.http_pool_connections,
            "http_pool_maxsize": self.test_grid_config.http_pool_maxsize,
            "http_request_timeout":
                self.test_grid_config.http_request_timeout,
            "admin_user": {"name": self.test_grid_config.admin_user_name,
                           "password": self.test_grid_config.admin_password}
        }
        actual_grid_connection = jsonutils.loads(grids[0]['grid_connection'])
        self.assertEqual(expected_grid_connection, actual_grid_connection)
        self.assertEqual('ON', grids[0]['grid_status'])

        # change active grid to "Test Grid 2"
        new_active_grid_config = grid.GridConfiguration(self.ctx)
        new_active_grid_config.gm_connector = mock.Mock()
        new_active_grid_config.grid_id = 200
        new_active_grid_config.grid_name = "Test Grid 2"
        new_active_grid_config.grid_master_host = '192.168.1.8'
        new_active_grid_config.admin_user_name = 'admin'
        new_active_grid_config.admin_password = 'infoblox'
        new_active_grid_config.wapi_version = '1.4.2'
        member_mgr = member.GridMemberManager(new_active_grid_config)
        member_mgr.sync_grid()

        grids = dbi.get_grids(self.ctx.session)
        self.assertEqual(2, len(grids))
        self.assertEqual(self.test_grid_config.grid_id, grids[0]['grid_id'])
        self.assertEqual('OFF', grids[0]['grid_status'])

        self.assertEqual(new_active_grid_config.grid_id, grids[1]['grid_id'])
        self.assertEqual(new_active_grid_config.grid_name,
                         grids[1]['grid_name'])
        expected_grid_connection = {
            "wapi_version": new_active_grid_config.wapi_version,
            "ssl_verify": new_active_grid_config.ssl_verify,
            "http_pool_connections":
                new_active_grid_config.http_pool_connections,
            "http_pool_maxsize": new_active_grid_config.http_pool_maxsize,
            "http_request_timeout":
                new_active_grid_config.http_request_timeout,
            "admin_user": {"name": new_active_grid_config.admin_user_name,
                           "password": new_active_grid_config.admin_password}
        }
        actual_grid_connection = jsonutils.loads(grids[1]['grid_connection'])
        self.assertEqual(expected_grid_connection, actual_grid_connection)
        self.assertEqual('ON', grids[1]['grid_status'])

    def _mock_member_mgr(self, member_mgr, discover_members=None,
                         discover_member_licenses=None):
        member_mgr._discover_members = mock.Mock(return_value=discover_members)
        member_mgr._discover_member_licenses = mock.Mock(
            return_value=discover_member_licenses)

        member_mgr._discover_dns_settings = mock.Mock(return_value=[])
        member_mgr._discover_dhcp_settings = mock.Mock(return_value=[])
        return member_mgr

    def test_sync_member_without_cloud_support(self):
        # wapi version less than 2.0 indicates no cloud support
        self.test_grid_config.wapi_version = '1.4.2'
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITHOUT_CLOUD)
        self._mock_member_mgr(member_mgr, discover_members=member_json)

        member_mgr.sync_members()

        members = dbi.get_members(self.ctx.session)
        self.assertEqual(len(member_json), len(members))
        for m in members:
            if self.test_grid_config.grid_master_host == m['member_ip']:
                self.assertEqual('GM', m['member_type'])
            else:
                self.assertEqual('REGULAR', m['member_type'])

    def test_sync_member_with_cloud_support_without_member_licenses(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITH_CLOUD)
        self._mock_member_mgr(member_mgr, discover_members=member_json)

        member_mgr.sync_members()

        members = dbi.get_members(self.ctx.session)
        self.assertEqual(len(member_json), len(members))
        for m in members:
            if self.test_grid_config.grid_master_host == m['member_ip']:
                self.assertEqual('GM', m['member_type'])
            else:
                self.assertEqual('REGULAR', m['member_type'])

    def test_sync_member_with_cloud_support_with_member_licenses(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr.sync_grid()

        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITH_CLOUD)
        license_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBER_LICENSES)
        self._mock_member_mgr(member_mgr, discover_members=member_json,
                              discover_member_licenses=license_json)

        member_mgr.sync_members()

        # member_type is computed base on member licenses. To prove correctness
        # of member types, we need to first figure out which member has
        # 'CLOUD_API' license. A member with 'CLOUD_API' is CPM; otherwise
        # 'REGULAR'. Grid master member is 'GM'
        cloud_member_list = []
        cloud_lecensed_members = utils.find_in_list('type', ['CLOUD_API'],
                                                    license_json)
        member_hwids = utils.get_values_from_records('hwid',
                                                     cloud_lecensed_members)
        for m in member_json:
            member_id_arg = str(self.test_grid_config.grid_id) + m['host_name']
            member_id = utils.get_hash(member_id_arg)
            member_hwid = m['node_info'][0].get('hwid')
            if member_hwid in member_hwids:
                cloud_member_list.append(member_id)

        # now we know member licenses so get members in db
        members = dbi.get_members(self.ctx.session)
        self.assertEqual(len(member_json), len(members))
        # verify member types
        for m in members:
            if self.test_grid_config.grid_master_host == m['member_ip']:
                self.assertEqual('GM', m['member_type'])
            else:
                if m['member_id'] in cloud_member_list:
                    self.assertEqual('CPM', m['member_type'])
                else:
                    self.assertEqual('REGULAR', m['member_type'])

    def test__discover_dns_settings(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_dns = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBER_DNS)

        return_fields = ['host_name', 'use_mgmt_port', 'use_mgmt_ipv6_port',
                         'use_lan_port', 'use_lan_ipv6_port', 'use_lan2_port',
                         'use_lan2_ipv6_port', 'additional_ip_list']

        get_mock = mock.Mock(return_value=member_dns)
        member_mgr._connector.get_object = get_mock

        dns_settings = member_mgr._discover_dns_settings()
        get_mock.assert_called_once_with('member:dns',
                                         return_fields=return_fields)
        self.assertEqual(1, len(dns_settings))
        self.assertTrue('nios-7.2.0-master.com' in dns_settings)

    def test__discover_dns_settings_old_wapi(self):
        # make sure wapi calls are not done on wapi older then 2.3
        self.test_grid_config.wapi_version = '2.2.1'
        member_mgr = member.GridMemberManager(self.test_grid_config)
        get_mock = mock.Mock()
        member_mgr._connector.get_object = get_mock

        dns_settings = member_mgr._discover_dns_settings()
        self.assertEqual({}, dns_settings)
        get_mock.assert_not_called()

    def test__discover_dhcp_settings(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_dhcp = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBER_DHCP)

        return_fields = ['host_name', 'enable_dhcp']

        get_mock = mock.Mock(return_value=member_dhcp)
        member_mgr._connector.get_object = get_mock

        dhcp_settings = member_mgr._discover_dhcp_settings()
        get_mock.assert_called_once_with('member:dhcpproperties',
                                         return_fields=return_fields)
        self.assertEqual(1, len(dhcp_settings))
        self.assertTrue('nios-7.2.0-master.com' in dhcp_settings)

    def test__discover_dhcp_settings_old_wapi(self):
        # make sure wapi calls are not done on wapi older then 2.2.1
        self.test_grid_config.wapi_version = '2.2'
        member_mgr = member.GridMemberManager(self.test_grid_config)
        get_mock = mock.Mock()
        member_mgr._connector.get_object = get_mock

        dhcp_settings = member_mgr._discover_dhcp_settings()
        self.assertEqual({}, dhcp_settings)
        get_mock.assert_not_called()

    def test__get_dhcp_ips_lan1(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_dict = {'host_name': 'nios-7.2.0-master.com',
                       'vip_setting': {'address': '172.22.0.10'},
                       'ipv6_setting': {'virtual_ip': '2001::12'}}
        dhcp_settings = {'nios-7.2.0-master.com':
                         {'host_name': 'nios-7.2.0-master.com',
                          'enabled': True}}
        # should return lan1 settings
        ip, ipv6 = member_mgr._get_dhcp_ips(member_dict, dhcp_settings)
        self.assertEqual('172.22.0.10', ip)
        self.assertEqual('2001::12', ipv6)

    def test__get_dhcp_ips_lan2(self):
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_dict = {'host_name': 'nios-7.2.0-master.com',
                       'vip_setting': {'address': '172.22.0.10'},
                       'ipv6_setting': {'virtual_ip': '2001::12'},
                       'lan2_port_setting':
                           {'network_setting': {'address': '172.25.0.10'},
                            'v6_network_setting': {'virtual_ip': '2022::25'}}}
        dhcp_settings = {'nios-7.2.0-master.com':
                         {'host_name': 'nios-7.2.0-master.com',
                          'enabled': False}}
        # should return lan2 settings
        ip, ipv6 = member_mgr._get_dhcp_ips(member_dict, dhcp_settings)
        self.assertEqual('172.25.0.10', ip)
        self.assertEqual('2022::25', ipv6)

    def _test__get_dns_ips(self, use_lan2_ipv6_port=False,
                           use_lan2_port=False, use_lan_ipv6_port=False,
                           use_lan_port=False, use_mgmt_ipv6_port=False,
                           use_mgmt_port=False, additional_ips=None):
        if additional_ips is None:
            additional_ips = []
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_dict = {'host_name': 'nios-7.2.0-master.com',
                       'vip_setting': {'address': '172.22.0.10'},
                       'ipv6_setting': {'virtual_ip': '2001::12'},
                       'node_info': [
                           {'v6_mgmt_network_setting': {
                               'virtual_ip': '2050::55'},
                            'mgmt_network_setting': {
                                'address': '192.168.1.85'}}],
                       'lan2_port_setting':
                           {'network_setting': {'address': '172.25.0.10'},
                            'v6_network_setting': {'virtual_ip': '2022::25'}}}
        dns_settings = {'nios-7.2.0-master.com':
                        {'host_name': 'nios-7.2.0-master.com',
                         "additional_ip_list": additional_ips,
                         "use_lan2_ipv6_port": use_lan2_ipv6_port,
                         "use_lan2_port": use_lan2_port,
                         "use_lan_ipv6_port": use_lan_ipv6_port,
                         "use_lan_port": use_lan_port,
                         "use_mgmt_ipv6_port": use_mgmt_ipv6_port,
                         "use_mgmt_port": use_mgmt_port}}
        return member_mgr._get_dns_ips(member_dict, dns_settings)

    def test__get_dns_ips_lan1(self):
        ip, ipv6 = self._test__get_dns_ips(use_lan_port=True,
                                           use_lan_ipv6_port=True)
        self.assertEqual('172.22.0.10', ip)
        self.assertEqual('2001::12', ipv6)

    def test__get_dns_ips_lan2(self):
        ip, ipv6 = self._test__get_dns_ips(use_lan2_port=True,
                                           use_lan2_ipv6_port=True)
        self.assertEqual('172.25.0.10', ip)
        self.assertEqual('2022::25', ipv6)

    def test__get_dns_ips_mgmt(self):
        ip, ipv6 = self._test__get_dns_ips(use_mgmt_port=True,
                                           use_mgmt_ipv6_port=True)
        self.assertEqual('192.168.1.85', ip)
        self.assertEqual('2050::55', ipv6)

    def test__get_dns_ips_additional_ips(self):
        additional_ips = ['145.22.0.15', '2012::125']
        ip, ipv6 = self._test__get_dns_ips(additional_ips=additional_ips)
        self.assertEqual('145.22.0.15', ip)
        self.assertEqual('2012::125', ipv6)

    def test__get_dns_ips_fallback_to_lan1(self):
        ip, ipv6 = self._test__get_dns_ips()
        self.assertEqual('172.22.0.10', ip)
        self.assertEqual('2001::12', ipv6)

    def _test_discover_members(self, wapi_version, return_fields):
        self.test_grid_config.wapi_version = wapi_version
        member_mgr = member.GridMemberManager(self.test_grid_config)
        member_mgr._connector = mock.Mock()

        member_mgr._discover_members()

        member_mgr._connector.get_object.assert_called_with(
            'member', return_fields=return_fields)

    def test__discover_members_all_fields(self):
        wapi_version = '2.2'
        return_fields = ['node_info', 'host_name', 'vip_setting', 'extattrs',
                         'ipv6_setting', 'lan2_port_setting',
                         'mgmt_port_setting']
        self._test_discover_members(wapi_version, return_fields)

    def test__discover_members_not_all_fields(self):
        wapi_version = '2.1'
        return_fields = ['node_info', 'host_name', 'vip_setting', 'extattrs']
        self._test_discover_members(wapi_version, return_fields)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests.unit import testlib_api
from neutron_lib import constants as n_const
from neutron_lib import context

from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import constants
from networking_infoblox.neutron.common import dns
from networking_infoblox.neutron.common import ea_manager
from networking_infoblox.neutron.db import infoblox_db as dbi
from networking_infoblox.tests import base


class DnsControllerTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):
        super(DnsControllerTestCase, self).setUp()
        self.neutron_cxt = context.get_admin_context()
        self.test_dns_zone = 'infoblox.com'
        self.ib_cxt = self._get_ib_context()
        self.ib_cxt.context = self.neutron_cxt
        self.ib_cxt.network_is_external = False
        self.ib_cxt.grid_config.zone_creation_strategy = (
            self._get_default_zone_creation_strategy())
        self.test_zone_format = "IPV%s" % self.ib_cxt.subnet['ip_version']
        self.controller = dns.DnsController(self.ib_cxt)
        self.controller.pattern_builder = mock.Mock()
        self.controller.pattern_builder.get_zone_name.return_value = (
            self.test_dns_zone)

    @staticmethod
    def _get_default_zone_creation_strategy():
        return constants.GRID_CONFIG_DEFAULTS[
            constants.EA_GRID_CONFIG_ZONE_CREATION_STRATEGY]

    def _get_ib_context(self):
        ib_cxt = mock.Mock()
        ib_cxt.network = {'id': 'network-id',
                          'name': 'test-net-1',
                          'tenant_id': 'network-id'}
        ib_cxt.subnet = {'id': 'subnet-id',
                         'name': 'test-sub-1',
                         'tenant_id': 'tenant-id',
                         'network_id': 'network-id',
                         'cidr': '11.11.1.0/24',
                         'ip_version': 4}
        ib_cxt.tenant_id = ib_cxt.network['tenant_id']
        ib_cxt.mapping.dns_view = 'test-dns-view'
        ib_cxt.mapping.network_view = 'test-network-view'
        ib_cxt.get_dns_members.return_value = ([mock.ANY], None)
        ib_cxt.grid_config.ns_group = None
        ib_cxt.grid_config.default_domain_name_pattern = self.test_dns_zone
        ib_cxt.grid_config.allow_static_zone_deletion = False
        return ib_cxt

    @mock.patch.object(ib_objects, 'DNSZone')
    def test_create_dns_zones_without_ns_group_zone_preexist(self,
                                                             mock_zone_obj):
        mock_zone_obj.create_check_exists.return_value = mock.Mock(), False
        self._test_create_dns_zones_without_ns_group_both_zones(mock_zone_obj)

    @mock.patch.object(ib_objects, 'DNSZone')
    def test_create_dns_zones_without_ns_group_zone_created(self,
                                                            mock_zone_obj):
        mock_zone_obj.create_check_exists.return_value = mock.Mock(), True
        self._test_create_dns_zones_without_ns_group_both_zones(mock_zone_obj)

    def _test_create_dns_zones_without_ns_group_both_zones(self,
                                                           mock_zone_obj):
        rollback_list = []
        # default strategy is to create both Forward and Reverse zones
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.test_dns_zone,
                grid_primary=[mock.ANY],
                grid_secondaries=None,
                extattrs=mock.ANY),
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.ib_cxt.subnet['cidr'],
                grid_primary=[mock.ANY],
                prefix=None,
                zone_format=self.test_zone_format,
                extattrs=mock.ANY)
        ]
        self.ib_cxt.ibom.reset_mock()
        mock_zone_obj.reset_mock()

        # check strategy with only Forward zone
        self.ib_cxt.grid_config.zone_creation_strategy = [
            constants.ZONE_CREATION_STRATEGY_FORWARD]
        self.controller._update_strategy_and_eas()
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.test_dns_zone,
                grid_primary=[mock.ANY],
                grid_secondaries=None,
                extattrs=mock.ANY),
        ]
        self.ib_cxt.ibom.reset_mock()
        mock_zone_obj.reset_mock()

        # check strategy with only Reverse zone
        self.ib_cxt.grid_config.zone_creation_strategy = [
            constants.ZONE_CREATION_STRATEGY_REVERSE]
        self.controller._update_strategy_and_eas()
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.ib_cxt.subnet['cidr'],
                grid_primary=[mock.ANY],
                prefix=None,
                zone_format=self.test_zone_format,
                extattrs=mock.ANY)
        ]
        self.ib_cxt.ibom.reset_mock()
        mock_zone_obj.reset_mock()

        # check empty strategy
        self.ib_cxt.grid_config.zone_creation_strategy = []
        self.controller._update_strategy_and_eas()
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == []

    @mock.patch.object(ib_objects, 'DNSZone')
    def test_create_dns_zones_with_ns_group_zone_preexist(self, mock_zone_obj):
        mock_zone_obj.create_check_exists.return_value = mock.Mock(), False
        self._test_create_dns_zones_with_ns_group(mock_zone_obj)

    @mock.patch.object(ib_objects, 'DNSZone')
    def test_create_dns_zones_with_ns_group_zone_created(self, mock_zone_obj):
        mock_zone_obj.create_check_exists.return_value = mock.Mock(), True
        self._test_create_dns_zones_with_ns_group(mock_zone_obj)

    def _test_create_dns_zones_with_ns_group(self, mock_zone_obj):
        rollback_list = []
        # default strategy is to create both Forward and Reverse zones
        self.ib_cxt.grid_config.ns_group = 'test-ns-group'
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.test_dns_zone,
                ns_group=self.ib_cxt.grid_config.ns_group,
                extattrs=mock.ANY),
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.ib_cxt.subnet['cidr'],
                prefix=None,
                zone_format=self.test_zone_format,
                extattrs=mock.ANY)
        ]
        self.ib_cxt.ibom.reset_mock()
        mock_zone_obj.reset_mock()
        # check strategy with only Forward zone
        self.ib_cxt.grid_config.zone_creation_strategy = [
            constants.ZONE_CREATION_STRATEGY_FORWARD]
        self.controller._update_strategy_and_eas()
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.test_dns_zone,
                ns_group=self.ib_cxt.grid_config.ns_group,
                extattrs=mock.ANY),
        ]
        self.ib_cxt.ibom.reset_mock()
        mock_zone_obj.reset_mock()

        # check strategy with only Reverse zone
        self.ib_cxt.grid_config.zone_creation_strategy = [
            constants.ZONE_CREATION_STRATEGY_REVERSE]
        self.controller._update_strategy_and_eas()
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=self.ib_cxt.mapping.dns_view,
                fqdn=self.ib_cxt.subnet['cidr'],
                prefix=None,
                zone_format=self.test_zone_format,
                extattrs=mock.ANY)
        ]

        # check empty strategy
        self.ib_cxt.ibom.reset_mock()
        mock_zone_obj.reset_mock()
        self.ib_cxt.grid_config.zone_creation_strategy = []
        self.controller._update_strategy_and_eas()
        self.controller.create_dns_zones(rollback_list)
        assert mock_zone_obj.method_calls == []

    def _create_ib_zone_ea(self):
        zone_ea = {'CMP Type': {'value': 'OpenStack'},
                   'Cloud API Owned': {'value': 'True'},
                   'Tenant ID': {'value': 'test-id'},
                   'Tenant Name': {'value': 'tenant-name'},
                   'Account': {'value': 'admin'},
                   'Network View ID': {'value': 'default'}}
        return ib_objects.EA.from_dict(zone_ea)

    def _reset_ib_zone_ea(self):
        expected_ea = {'Cloud API Owned': {'value': 'False'},
                       'CMP Type': {'value': 'N/A'},
                       'Tenant ID': {'value': 'N/A'}}
        return expected_ea

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    def test_delete_dns_zones_for_shared_network_view(self):
        self.ib_cxt.mapping.shared = True
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = False

        ib_zone_ea = self._create_ib_zone_ea()
        ib_zone_mock = mock.Mock(extattrs=ib_zone_ea)

        with mock.patch.object(ib_objects.DNSZone,
                               'search',
                               return_value=ib_zone_mock):
            self.controller.delete_dns_zones()
            assert ib_zone_mock.extattrs.to_dict() == ib_zone_ea.to_dict()

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    def test_delete_dns_zones_for_external_network(self):
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = True
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = False

        ib_zone_ea = self._create_ib_zone_ea()
        ib_zone_mock = mock.Mock(extattrs=ib_zone_ea)

        with mock.patch.object(ib_objects.DNSZone,
                               'search',
                               return_value=ib_zone_mock):
            self.controller.delete_dns_zones()
            assert ib_zone_mock.extattrs.to_dict() == ib_zone_ea.to_dict()

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_private_networks', mock.Mock())
    def test_delete_dns_zones_for_shared_network(self):
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = True
        self.ib_cxt.grid_config.admin_network_deletion = False

        ib_zone_ea = self._create_ib_zone_ea()
        ib_zone_mock = mock.Mock(extattrs=ib_zone_ea)

        with mock.patch.object(ib_objects.DNSZone,
                               'search',
                               return_value=ib_zone_mock):
            self.controller.delete_dns_zones()
            assert ib_zone_mock.extattrs.to_dict() == ib_zone_ea.to_dict()

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_private_networks', mock.Mock())
    def test_delete_dns_zones_for_shared_network_with_admin_network_deletable(
            self):
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = True
        self.ib_cxt.grid_config.admin_network_deletion = True

        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        dbi.is_last_subnet_in_private_networks.assert_not_called()

        # Now enable static zone deletion
        self.ib_cxt.ibom.reset_mock()
        dbi.is_last_subnet_in_private_networks.reset_mock()
        self.ib_cxt.grid_config.allow_static_zone_deletion = True
        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.test_dns_zone),
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        assert dbi.is_last_subnet_in_private_networks.called

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_private_networks', mock.Mock())
    def test_delete_dns_zones_for_private_network_with_static_zone(self):
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = True

        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        dbi.is_last_subnet_in_private_networks.assert_not_called()

        # Now enable static zone deletion
        self.ib_cxt.ibom.reset_mock()
        dbi.is_last_subnet_in_private_networks.reset_mock()
        self.ib_cxt.grid_config.allow_static_zone_deletion = True
        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.test_dns_zone),
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        assert dbi.is_last_subnet_in_private_networks.called

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_private_networks', mock.Mock())
    def test_delete_dns_zones_for_private_network_with_subnet_pattern(self):
        self.ib_cxt.grid_config.default_domain_name_pattern = (
            '{subnet_name}.infoblox.com')
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = False
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False

        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.test_dns_zone),
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        assert not dbi.is_last_subnet_in_private_networks.called

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_network', mock.Mock())
    def test_delete_dns_zones_for_private_network_with_network_pattern(self):
        self.ib_cxt.grid_config.default_domain_name_pattern = (
            '{network_id}.infoblox.com')
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = False
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False

        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.test_dns_zone),
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        assert dbi.is_last_subnet_in_network.called

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_tenant', mock.Mock())
    def test_delete_dns_zones_for_private_network_with_tenant_pattern(self):
        self.ib_cxt.grid_config.default_domain_name_pattern = (
            '{tenant_name}.infoblox.com')
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = False
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False

        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.test_dns_zone),
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        assert dbi.is_last_subnet_in_tenant.called

    @mock.patch.object(dbi, 'get_network_views', mock.Mock())
    @mock.patch.object(dbi, 'is_last_subnet_in_address_scope', mock.Mock())
    def test_delete_dns_zones_for_private_network_with_address_scope_pattern(
            self):
        self.ib_cxt.grid_config.default_domain_name_pattern = (
            '{address_scope_id}.infoblox.com')
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network['router:external'] = False
        self.ib_cxt.network['shared'] = False
        self.ib_cxt.grid_config.admin_network_deletion = False
        self.ib_cxt.mapping.shared = False
        self.ib_cxt.network_is_shared_or_external = False

        self.controller.delete_dns_zones()

        assert self.ib_cxt.ibom.method_calls == [
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.test_dns_zone),
            mock.call.delete_dns_zone(
                self.ib_cxt.mapping.dns_view,
                self.ib_cxt.subnet['cidr'])
        ]
        assert dbi.is_last_subnet_in_address_scope.called

    @mock.patch.object(dbi, 'get_instance', mock.Mock())
    @mock.patch.object(ea_manager, 'get_ea_for_ip', mock.Mock())
    @mock.patch.object(dns.DnsController, '_ensure_dns_zone_availability',
                       mock.Mock())
    def test_bind_names(self):
        ip_address = '11.11.1.2'
        instance_name = 'test-vm'
        tenant_id = 'tenant-id'
        tenant_name = 'tenant-name'
        port_id = 'port-id'
        port_name = 'port-name'
        device_id = 'device-id'
        self.ib_cxt.user_id = 'test-id'
        self.ib_cxt.mapping.network_view = 'test-view'
        self.ib_cxt.get_tenant_name.return_value = tenant_name
        instance = mock.MagicMock()
        instance.instance_name = instance_name
        dbi.get_instance.return_value = instance

        fqdn = str.format("{}.{}", instance_name, self.test_dns_zone)
        self.controller.pattern_builder.get_hostname.return_value = fqdn

        ip_alloc_method_calls = [
            mock.call.bind_names(self.ib_cxt.mapping.network_view,
                                 self.ib_cxt.mapping.dns_view,
                                 ip_address,
                                 fqdn,
                                 mock.ANY)
        ]
        ea_call_params = [
            self.ib_cxt.user_id, tenant_id, tenant_name, self.ib_cxt.network,
            port_id, device_id, 'device_owner', False, instance_name]
        test_data = [
            {
                'device_owner': None,
                'instance_name': instance_name,
                'ip_alloc': self.ib_cxt.ip_alloc,
                'ip_alloc_calls': [],
                'dbi_call_params': None,
            },
            {
                'device_owner': n_const.DEVICE_OWNER_DHCP,
                'instance_name': instance_name,
                'ip_alloc': self.ib_cxt.dhcp_port_ip_alloc,
                'ip_alloc_calls': ip_alloc_method_calls,
                'dbi_call_params': None,
            },
            {
                'device_owner': n_const.DEVICE_OWNER_ROUTER_GW,
                'instance_name': instance_name,
                'ip_alloc': self.ib_cxt.ip_alloc,
                'ip_alloc_calls': ip_alloc_method_calls,
                'dbi_call_params': None,
            },
            {
                'device_owner': constants.NEUTRON_DEVICE_OWNER_COMPUTE_NOVA,
                'instance_name': instance_name,
                'ip_alloc': self.ib_cxt.ip_alloc,
                'ip_alloc_calls': ip_alloc_method_calls,
                'dbi_call_params': None,
            },
            {
                'device_owner': constants.NEUTRON_DEVICE_OWNER_COMPUTE_NOVA,
                'instance_name': None,
                'ip_alloc': self.ib_cxt.ip_alloc,
                'ip_alloc_calls': ip_alloc_method_calls,
                'dbi_call_params': [self.ib_cxt.context.session, device_id],
            },
        ]

        for data in test_data:
            self.controller.bind_names(
                ip_address, data['instance_name'], port_id,
                port_tenant_id=tenant_id, device_id=device_id,
                device_owner=data['device_owner'], port_name=port_name)
            ip_alloc = data['ip_alloc']
            assert ip_alloc.method_calls == data['ip_alloc_calls']
            ip_alloc.reset_mock()
            if data['device_owner'] is None:
                ea_manager.get_ea_for_ip.assert_not_called()
            else:
                ea_call_params[6] = data['device_owner']
                ea_manager.get_ea_for_ip.assert_called_once_with(
                    *ea_call_params)
                ea_manager.get_ea_for_ip.reset_mock()
            dbi_call_params = data['dbi_call_params']
            if dbi_call_params is None:
                dbi.get_instance.assert_not_called()
            else:
                dbi.get_instance.assert_called_once_with(*dbi_call_params)
                ea_manager.get_ea_for_ip.reset_mock()

    @mock.patch.object(dns.DnsController, '_ensure_dns_zone_availability',
                       mock.Mock())
    def test_unbind_names(self):
        ip_address = '11.11.1.2'
        instance_name = 'test-vm'
        port_id = 'port-id'
        port_name = 'port-name'

        fqdn = str.format("{}.{}", instance_name, self.test_dns_zone)
        self.controller.pattern_builder.get_hostname.return_value = fqdn

        self.controller.unbind_names(ip_address, instance_name, port_id,
                                     device_owner=None, port_name=port_name)
        assert self.ib_cxt.ip_alloc.method_calls == []

        self.controller.unbind_names(ip_address, instance_name, port_id,
                                     device_owner=n_const.DEVICE_OWNER_DHCP,
                                     port_name=port_name)
        assert self.ib_cxt.dhcp_port_ip_alloc.method_calls == [
            mock.call.unbind_names(self.ib_cxt.mapping.network_view,
                                   self.ib_cxt.mapping.dns_view,
                                   ip_address,
                                   fqdn,
                                   None)
        ]

        self.controller.unbind_names(
            ip_address, instance_name, port_id,
            device_owner=n_const.DEVICE_OWNER_ROUTER_GW, port_name=port_name)
        assert self.ib_cxt.ip_alloc.method_calls == [
            mock.call.unbind_names(self.ib_cxt.mapping.network_view,
                                   self.ib_cxt.mapping.dns_view,
                                   ip_address,
                                   fqdn,
                                   None)
        ]

    @mock.patch.object(dns.DnsController, '_ensure_dns_zone_availability',
                       mock.Mock())
    def test_unbind_names_without_name(self):
        ip_address = '11.11.1.2'
        port_id = 'port-id'
        port_name = 'port-name'
        self.ib_cxt.grid_config.default_host_name_pattern = '{instance_name}'
        self.ib_cxt.grid_config.default_domain_name_pattern = '{subnet_id}.com'
        controller = dns.DnsController(self.ib_cxt)
        controller.unbind_names(ip_address, None, port_id,
                                device_owner='compute:nova',
                                port_name=port_name)
        assert self.ib_cxt.ip_alloc.method_calls == [
            mock.call.unbind_names(
                self.ib_cxt.mapping.network_view, self.ib_cxt.mapping.dns_view,
                ip_address, None, None)]

    def _test_update_calls(self, strategy, calls):
        self.ib_cxt.grid_config.zone_creation_strategy = strategy
        self.controller._update_strategy_and_eas()
        self.controller.update_dns_zones()
        assert self.ib_cxt.ibom.method_calls == calls
        self.ib_cxt.ibom.reset_mock()

    def test_update_dns_zones_all(self):
        self._test_update_calls(
            self._get_default_zone_creation_strategy(),
            [
                mock.call.update_dns_zone_attrs(
                    self.ib_cxt.mapping.dns_view,
                    self.test_dns_zone,
                    mock.ANY),
                mock.call.update_dns_zone_attrs(
                    self.ib_cxt.mapping.dns_view,
                    self.ib_cxt.subnet['cidr'],
                    mock.ANY)
            ])

    def test_update_dns_zones_forward(self):
        self._test_update_calls(
            [constants.ZONE_CREATION_STRATEGY_FORWARD],
            [
                mock.call.update_dns_zone_attrs(
                    self.ib_cxt.mapping.dns_view,
                    self.test_dns_zone,
                    mock.ANY),
            ])

    def test_update_dns_zones_reverse(self):
        self._test_update_calls(
            [constants.ZONE_CREATION_STRATEGY_REVERSE],
            [
                mock.call.update_dns_zone_attrs(
                    self.ib_cxt.mapping.dns_view,
                    self.ib_cxt.subnet['cidr'],
                    mock.ANY)
            ])

    def test_update_dns_zones_empty(self):
        self._test_update_calls([], [])

    @mock.patch.object(ib_objects, 'DNSZone')
    @mock.patch.object(ea_manager, 'get_ea_for_forward_zone')
    def test_ensure_dns_zone_with_ns_group_availability(
            self, mock_eas_fw_zone, mock_zone_obj):
        # default strategy is to create both Forward
        self.ib_cxt.grid_config.ns_group = 'test-ns-group'
        dns_view = 'test_dns_view'
        port_tenant_id = 'test_tenant_id'
        tenant_name = 'test_tenant_name'
        is_external = 'false'
        mock_zone_obj.create_check_exists.return_value = (
            mock.Mock(), True)
        mock_eas_fw_zone.return_value = 'test_eas_forward_zone'
        self.controller._ensure_dns_zone_availability(
            dns_view, port_tenant_id,
            tenant_name, is_external)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=dns_view,
                fqdn=self.test_dns_zone,
                ns_group=self.ib_cxt.grid_config.ns_group,
                extattrs='test_eas_forward_zone')
        ]

    @mock.patch.object(ib_objects, 'DNSZone')
    @mock.patch.object(ea_manager, 'get_ea_for_forward_zone')
    def test_ensure_dns_zone_without_ns_group_availability(
            self, mock_eas_fw_zone, mock_zone_obj):
        dns_view = 'test_dns_view'
        port_tenant_id = 'test_tenant_id'
        tenant_name = 'test_tenant_name'
        is_external = 'false'
        mock_zone_obj.create_check_exists.return_value = (
            mock.Mock(), True)
        mock_eas_fw_zone.return_value = 'test_eas_forward_zone'
        self.controller._ensure_dns_zone_availability(
            dns_view, port_tenant_id,
            tenant_name, is_external)
        assert mock_zone_obj.method_calls == [
            mock.call.create_check_exists(
                self.ib_cxt.connector,
                view=dns_view,
                fqdn=self.test_dns_zone,
                grid_primary=[mock.ANY],
                grid_secondaries=None,
                extattrs='test_eas_forward_zone')
        ]



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import random
from sqlalchemy import func
from sqlalchemy.sql.expression import true

from neutron.db.models import address_scope as address_scope_db
from neutron.db.models import external_net as external_net_db
from neutron.db.models import l3 as l3_db
from neutron.db import models_v2

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.db import infoblox_models as ib_models


# Grid Management
def get_grids(session, grid_id=None, grid_name=None, grid_status=None):
    q = session.query(ib_models.InfobloxGrid)
    if grid_id:
        q = q.filter(ib_models.InfobloxGrid.grid_id == grid_id)
    if grid_name:
        q = q.filter(ib_models.InfobloxGrid.grid_name == grid_name)
    if grid_status:
        q = q.filter(ib_models.InfobloxGrid.grid_status == grid_status)
    return q.all()


def add_grid(session, grid_id, grid_name, grid_connection, grid_status, gm_id):
    grid = ib_models.InfobloxGrid(grid_id=grid_id,
                                  grid_name=grid_name,
                                  grid_connection=grid_connection,
                                  grid_status=grid_status,
                                  gm_id=gm_id)
    session.add(grid)
    return grid


def update_grid(session, grid_id, grid_name=None, grid_connection=None,
                grid_status=None):
    update_data = dict()
    if grid_name:
        update_data['grid_name'] = grid_name
    if grid_connection:
        update_data['grid_connection'] = grid_connection
    if grid_status:
        update_data['grid_status'] = grid_status

    if update_data:
        session.query(ib_models.InfobloxGrid).\
            filter_by(grid_id=grid_id).\
            update(update_data)


def remove_grids(session, grid_ids):
    if grid_ids and isinstance(grid_ids, list):
        with session.begin(subtransactions=True):
            q = session.query(ib_models.InfobloxGrid)
            q = q.filter(ib_models.InfobloxGrid.grid_id.in_(grid_ids))
            q.delete(synchronize_session=False)


# Grid Members
def get_members(session, member_id=None, grid_id=None, member_name=None,
                member_type=None, member_status=None):
    q = session.query(ib_models.InfobloxGridMember)
    if member_id:
        q = q.filter(ib_models.InfobloxGridMember.member_id == member_id)
    if grid_id:
        q = q.filter(ib_models.InfobloxGridMember.grid_id == grid_id)
    if member_name:
        q = q.filter(ib_models.InfobloxGridMember.member_name == member_name)
    if member_type:
        q = q.filter(ib_models.InfobloxGridMember.member_type == member_type)
    if member_status:
        q = q.filter(ib_models.InfobloxGridMember.member_status ==
                     member_status)
    return q.all()


def search_members(session, member_ids=None, member_names=None,
                   member_types=None, member_statuses=None):
    q = session.query(ib_models.InfobloxGridMember)
    if member_ids and isinstance(member_ids, list):
        q = q.filter(ib_models.InfobloxGridMember.member_id.in_(member_ids))
    if member_names and isinstance(member_names, list):
        q = q.filter(ib_models.InfobloxGridMember.member_name.
                     in_(member_names))
    if member_types and isinstance(member_types, list):
        q = q.filter(ib_models.InfobloxGridMember.member_type.
                     in_(member_types))
    if member_statuses and isinstance(member_statuses, list):
        q = q.filter(ib_models.InfobloxGridMember.member_status.
                     in_(member_statuses))
    return q.all()


def add_member(session, member_id, grid_id, member_name, member_ip,
               member_ipv6, member_type, member_status, member_dhcp_ip,
               member_dhcp_ipv6, member_dns_ip, member_dns_ipv6,
               member_wapi):
    member = ib_models.InfobloxGridMember(member_id=member_id,
                                          grid_id=grid_id,
                                          member_name=member_name,
                                          member_ip=member_ip,
                                          member_ipv6=member_ipv6,
                                          member_type=member_type,
                                          member_status=member_status,
                                          member_dhcp_ip=member_dhcp_ip,
                                          member_dhcp_ipv6=member_dhcp_ipv6,
                                          member_dns_ip=member_dns_ip,
                                          member_dns_ipv6=member_dns_ipv6,
                                          member_wapi=member_wapi)
    session.add(member)
    return member


def update_member(session, member_id, grid_id, member_name=None,
                  member_ip=None, member_ipv6=None,
                  member_type=None, member_status=None,
                  member_dhcp_ip=None, member_dhcp_ipv6=None,
                  member_dns_ip=None, member_dns_ipv6=None,
                  member_wapi=None):
    update_data = dict()
    if member_name:
        update_data['member_name'] = member_name
    if member_ip:
        update_data['member_ip'] = member_ip
    if member_ipv6:
        update_data['member_ipv6'] = member_ipv6
    if member_type:
        update_data['member_type'] = member_type
    if member_status:
        update_data['member_status'] = member_status
    if member_dhcp_ip:
        update_data['member_dhcp_ip'] = member_dhcp_ip
    if member_dhcp_ipv6:
        update_data['member_dhcp_ipv6'] = member_dhcp_ipv6
    if member_dns_ip:
        update_data['member_dns_ip'] = member_dns_ip
    if member_dns_ipv6:
        update_data['member_dns_ipv6'] = member_dns_ipv6
    if member_wapi:
        update_data['member_wapi'] = member_wapi

    if update_data:
        session.query(ib_models.InfobloxGridMember).\
            filter_by(member_id=member_id, grid_id=grid_id).\
            update(update_data)


def remove_members(session, member_ids):
    if member_ids and isinstance(member_ids, list):
        with session.begin(subtransactions=True):
            q = session.query(ib_models.InfobloxGridMember)
            q = q.filter(ib_models.InfobloxGridMember.member_id.in_(
                         member_ids))
            q.delete(synchronize_session=False)


# Network Views
def get_network_views(session, network_view_id=None, network_view=None,
                      grid_id=None, authority_member_id=None, shared=None,
                      dns_view=None, internal_network_view=None,
                      internal_dns_view=None, participated=None):
    q = session.query(ib_models.InfobloxNetworkView)
    if network_view_id:
        q = q.filter(ib_models.InfobloxNetworkView.id == network_view_id)
    if network_view:
        q = q.filter(ib_models.InfobloxNetworkView.network_view ==
                     network_view)
    if authority_member_id:
        q = q.filter(ib_models.InfobloxNetworkView.authority_member_id ==
                     authority_member_id)
    if shared:
        q = q.filter(ib_models.InfobloxNetworkView.shared == shared)
    if participated:
        q = q.filter(ib_models.InfobloxNetworkView.participated ==
                     participated)
    if dns_view:
        q = q.filter(ib_models.InfobloxNetworkView.dns_view == dns_view)
    if internal_network_view:
        q = q.filter(ib_models.InfobloxNetworkView.internal_network_view ==
                     internal_network_view)
    if internal_dns_view:
        q = q.filter(ib_models.InfobloxNetworkView.internal_dns_view ==
                     internal_dns_view)
    if grid_id:
        q = q.filter(ib_models.InfobloxNetworkView.grid_id == grid_id)
    return q.all()


def get_network_view_by_mapping(session, network_view_id=None, grid_id=None,
                                network_id=None, subnet_id=None):
    netview_mapping = get_network_view_mappings(
        session,
        network_view_id=network_view_id,
        network_id=network_id,
        subnet_id=subnet_id)
    if not netview_mapping:
        return None

    if len(netview_mapping) > 1:
        raise exc.MultipleNetworkViewMappingFound()

    netview_id = netview_mapping[0].network_view_id
    netviews = get_network_views(
        session,
        network_view_id=netview_id,
        grid_id=grid_id,
        participated=True)
    return netviews


def update_network_view(session, network_view_id, network_view,
                        authority_member_id, shared, dns_view, participated,
                        is_default):
    (session.query(ib_models.InfobloxNetworkView).
     filter_by(id=network_view_id).
     update({'network_view': network_view,
             'authority_member_id': authority_member_id,
             'shared': shared,
             'dns_view': dns_view,
             'participated': participated,
             'default': is_default}))


def update_network_view_id(session, old_id, new_id):
    (session.query(ib_models.InfobloxNetworkView).
     filter_by(id=old_id).
     update({'id': new_id}))


def add_network_view(session, network_view_id, network_view, grid_id,
                     authority_member_id, shared, dns_view,
                     internal_network_view, internal_dns_view, participated,
                     is_default):
    network_view = ib_models.InfobloxNetworkView(
        id=network_view_id,
        network_view=network_view,
        grid_id=grid_id,
        authority_member_id=authority_member_id,
        shared=shared,
        dns_view=dns_view,
        internal_network_view=internal_network_view,
        internal_dns_view=internal_dns_view,
        participated=participated,
        default=is_default)
    session.add(network_view)
    session.flush()
    return network_view


def remove_network_views(session, ids):
    if ids and isinstance(ids, list):
        with session.begin(subtransactions=True):
            q = session.query(ib_models.InfobloxNetworkView)
            q = q.filter(ib_models.InfobloxNetworkView.id.in_(ids))
            q.delete(synchronize_session=False)


def remove_network_views_by_names(session, network_views, grid_id):
    if network_views and isinstance(network_views, list):
        with session.begin(subtransactions=True):
            q = session.query(ib_models.InfobloxNetworkView)
            q = q.filter(ib_models.InfobloxNetworkView.grid_id == grid_id)
            q = q.filter(ib_models.InfobloxNetworkView.network_view.in_(
                         network_views))
            q.delete(synchronize_session=False)


# Network View Mapped to Neutron
def get_network_view_mappings(session, network_view_id=None, network_id=None,
                              subnet_id=None):
    q = session.query(ib_models.InfobloxNetworkViewMapping)
    if network_view_id:
        q = q.filter(ib_models.InfobloxNetworkViewMapping.network_view_id ==
                     network_view_id)
    if network_id:
        q = q.filter(ib_models.InfobloxNetworkViewMapping.network_id ==
                     network_id)
    if subnet_id:
        q = q.filter(ib_models.InfobloxNetworkViewMapping.subnet_id ==
                     subnet_id)

    sub_qry = session.query(ib_models.InfobloxNetworkView.id)
    sub_qry = sub_qry.filter(ib_models.InfobloxNetworkView.participated ==
                             true())
    q = q.filter(ib_models.InfobloxNetworkViewMapping.network_view_id.in_(
                 sub_qry))
    return q.all()


def associate_network_view(session, network_view_id, network_id, subnet_id):
    # check if network and subnet level mapping exists
    q = session.query(ib_models.InfobloxNetworkViewMapping)
    network_view_mapping = q.filter_by(network_id=network_id,
                                       subnet_id=subnet_id).first()
    if not network_view_mapping:
        network_view_mapping = ib_models.InfobloxNetworkViewMapping(
            network_view_id=network_view_id,
            network_id=network_id,
            subnet_id=subnet_id)
        session.add(network_view_mapping)


def dissociate_network_view(session, network_id, subnet_id):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxNetworkViewMapping)
        q = q.filter_by(network_id=network_id, subnet_id=subnet_id)
        q.delete(synchronize_session=False)


# Mapping Conditions
def get_mapping_conditions(session, network_view_id=None, grid_id=None,
                           neutron_object_name=None,
                           neutron_object_value=None):
    q = session.query(ib_models.InfobloxMappingCondition)
    if network_view_id:
        q = q.filter(ib_models.InfobloxMappingCondition.network_view_id ==
                     network_view_id)
    if neutron_object_name:
        q = q.filter(ib_models.InfobloxMappingCondition.neutron_object_name ==
                     neutron_object_name)
    if neutron_object_value:
        q = q.filter(ib_models.InfobloxMappingCondition.neutron_object_value ==
                     neutron_object_value)
    if grid_id:
        sub_qry = session.query(ib_models.InfobloxNetworkView.id)
        sub_qry = sub_qry.filter(ib_models.InfobloxNetworkView.grid_id ==
                                 grid_id)
        q = q.filter(ib_models.InfobloxMappingCondition.network_view_id.in_(
                     sub_qry))
    return q.all()


def add_mapping_condition(session, network_view_id, neutron_object_name,
                          neutron_object_value):
    mapping_condition = ib_models.InfobloxMappingCondition(
        network_view_id=network_view_id,
        neutron_object_name=neutron_object_name,
        neutron_object_value=neutron_object_value)
    session.add(mapping_condition)
    return mapping_condition


def add_mapping_conditions(session, network_view_id, neutron_object_name,
                           neutron_object_values):
    mapping_conditions = []
    if neutron_object_values and isinstance(neutron_object_values, list):
        for value in neutron_object_values:
            mapping_condition = ib_models.InfobloxMappingCondition(
                network_view_id=network_view_id,
                neutron_object_name=neutron_object_name,
                neutron_object_value=value
            )
            session.add(mapping_condition)
            mapping_conditions.append(mapping_condition)
    return mapping_conditions


def remove_mapping_condition(session, network_view_id, neutron_object_name,
                             neutron_object_value):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxMappingCondition)
        q = q.filter(ib_models.InfobloxMappingCondition.network_view_id ==
                     network_view_id)
        q = q.filter(ib_models.InfobloxMappingCondition.neutron_object_name ==
                     neutron_object_name)
        q = q.filter(ib_models.InfobloxMappingCondition.neutron_object_value ==
                     neutron_object_value)
        q.delete(synchronize_session=False)


def remove_mapping_conditions(session, network_view_id, neutron_object_name,
                              neutron_object_values):
    if network_view_id:
        with session.begin(subtransactions=True):
            q = session.query(ib_models.InfobloxMappingCondition)
            q = q.filter(ib_models.InfobloxMappingCondition.network_view_id ==
                         network_view_id)
            if neutron_object_name:
                q = q.filter(ib_models.InfobloxMappingCondition.
                             neutron_object_name == neutron_object_name)
            if (neutron_object_values and
                    isinstance(neutron_object_values, list)):
                q = q.filter(ib_models.InfobloxMappingCondition.
                             neutron_object_value.in_(neutron_object_values))
            q.delete(synchronize_session=False)


# Mapping Members
def get_mapping_members(session, network_view_id=None, member_id=None,
                        grid_id=None, mapping_relation=None):
    q = session.query(ib_models.InfobloxMappingMember)
    if network_view_id:
        q = q.filter(ib_models.InfobloxMappingMember.network_view_id ==
                     network_view_id)
    if member_id:
        q = q.filter(ib_models.InfobloxMappingMember.member_id == member_id)
    if mapping_relation:
        q = q.filter(ib_models.InfobloxMappingMember.mapping_relation ==
                     mapping_relation)
    if grid_id:
        sub_qry = session.query(ib_models.InfobloxNetworkView.id)
        sub_qry = sub_qry.filter(ib_models.InfobloxNetworkView.grid_id ==
                                 grid_id)
        q = q.filter(ib_models.InfobloxMappingMember.network_view_id.in_(
                     sub_qry))
    return q.all()


def add_mapping_member(session, network_view_id, member_id, mapping_relation):
    mapping_member = ib_models.InfobloxMappingMember(
        network_view_id=network_view_id,
        member_id=member_id,
        mapping_relation=mapping_relation)
    session.add(mapping_member)
    return mapping_member


def update_mapping_member(session, network_view_id, member_id,
                          mapping_relation):
    session.query(ib_models.InfobloxMappingMember).\
        filter_by(network_view_id=network_view_id, member_id=member_id).\
        update({'mapping_relation': mapping_relation})


def remove_mapping_member(session, network_view_id, member_id):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxMappingMember)
        q = q.filter_by(network_view_id=network_view_id, member_id=member_id)
        q.delete(synchronize_session=False)


# Member Reservation
def get_next_authority_member_for_ipam(session, grid_id):
    q = (session.query(
         ib_models.InfobloxGridMember,
         ib_models.InfobloxGridMember.member_id,
         func.count(ib_models.InfobloxNetworkView.id).label('count')).
         outerjoin(ib_models.InfobloxNetworkView,
                   ib_models.InfobloxNetworkView.authority_member_id ==
                   ib_models.InfobloxGridMember.member_id).
         filter(ib_models.InfobloxGridMember.grid_id == grid_id,
                ib_models.InfobloxGridMember.member_status ==
                const.MEMBER_STATUS_ON,
                ib_models.InfobloxGridMember.member_type !=
                const.MEMBER_TYPE_REGULAR_MEMBER).
         group_by(ib_models.InfobloxNetworkView.authority_member_id).
         order_by('count'))
    res = q.first()
    authority_member = res[0]
    return authority_member


def get_next_authority_member_for_dhcp(session, grid_id):
    q = (session.query(ib_models.InfobloxGridMember).
         outerjoin(ib_models.InfobloxMappingMember,
                   ib_models.InfobloxMappingMember.member_id ==
                   ib_models.InfobloxGridMember.member_id).
         outerjoin(ib_models.InfobloxServiceMember,
                   ib_models.InfobloxServiceMember.member_id ==
                   ib_models.InfobloxGridMember.member_id).
         filter(ib_models.InfobloxGridMember.grid_id == grid_id,
                ib_models.InfobloxGridMember.member_status ==
                const.MEMBER_STATUS_ON,
                ib_models.InfobloxGridMember.member_type ==
                const.MEMBER_TYPE_CP_MEMBER,
                ib_models.InfobloxMappingMember.member_id.is_(None),
                ib_models.InfobloxServiceMember.member_id.is_(None)))
    row_count = int(q.count())
    q = q.offset(int(row_count * random.random()))
    authority_member = q.first()
    return authority_member


def get_next_dhcp_member(session, grid_id, use_gm=True):
    """Get a next available dhcp member.

    For dhcp member, any member can be chosen but the priority is given to
    REGUALR first, then CPM. However gm can be omitted from selection
    if 'use_gm' parameter is set to False.
    """
    q = (session.query(ib_models.InfobloxGridMember).
         outerjoin(ib_models.InfobloxMappingMember,
                   ib_models.InfobloxMappingMember.member_id ==
                   ib_models.InfobloxGridMember.member_id).
         outerjoin(ib_models.InfobloxServiceMember,
                   ib_models.InfobloxServiceMember.member_id ==
                   ib_models.InfobloxGridMember.member_id).
         filter(ib_models.InfobloxGridMember.grid_id == grid_id,
                ib_models.InfobloxGridMember.member_status ==
                const.MEMBER_STATUS_ON,
                ib_models.InfobloxMappingMember.member_id.is_(None),
                ib_models.InfobloxServiceMember.member_id.is_(None)))
    if use_gm is False:
        q = q.filter(ib_models.InfobloxGridMember.member_type !=
                     const.MEMBER_TYPE_GRID_MASTER)
    q = q.order_by(ib_models.InfobloxGridMember.member_type.desc())
    row_count = int(q.count())
    q = q.offset(int(row_count * random.random()))
    dhcp_member = q.first()
    return dhcp_member


# Service Member Management
def get_service_members(session, network_view_id=None, member_id=None,
                        grid_id=None, service=None):
    q = session.query(ib_models.InfobloxServiceMember)
    if network_view_id:
        q = q.filter(ib_models.InfobloxServiceMember.network_view_id ==
                     network_view_id)
    if member_id:
        q = q.filter(ib_models.InfobloxServiceMember.member_id == member_id)
    if service:
        q = q.filter(ib_models.InfobloxServiceMember.service == service)
    if grid_id:
        sub_qry = (session.query(ib_models.InfobloxGridMember.member_id).
                   filter(ib_models.InfobloxGridMember.grid_id == grid_id))
        q = q.filter(ib_models.InfobloxServiceMember.member_id.in_(sub_qry))
    return q.all()


def add_service_member(session, network_view_id, member_id, service):
    service_member = ib_models.InfobloxServiceMember(
        network_view_id=network_view_id,
        member_id=member_id,
        service=service)
    session.add(service_member)
    return service_member


def remove_service_member(session, network_view_id, member_id=None,
                          service=None):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxServiceMember)
        q = q.filter(ib_models.InfobloxServiceMember.network_view_id ==
                     network_view_id)
        if member_id:
            q = q.filter(ib_models.InfobloxServiceMember.member_id ==
                         member_id)
        if service:
            q = q.filter(ib_models.InfobloxServiceMember.service == service)
        q.delete(synchronize_session=False)


def remove_service_members(session, network_view_id, member_ids):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxServiceMember)
        q = q.filter(ib_models.InfobloxServiceMember.network_view_id ==
                     network_view_id)
        if member_ids and isinstance(member_ids, list):
            q = q.filter(ib_models.InfobloxServiceMember.member_id.in_(
                         member_ids))
        q.delete(synchronize_session=False)


# Operational Setting Management
def add_operation_type(session, op_type, op_value):
    operation = ib_models.InfobloxOperation(
        op_type=op_type,
        op_value=op_value)
    session.add(operation)
    return operation


def get_last_sync_time(session):
    q = session.query(ib_models.InfobloxOperation)
    op_row = q.filter_by(op_type='last_sync_time').first()
    if op_row is None:
        add_operation_type(session, op_type='last_sync_time', op_value='')
        return None
    if op_row.op_value == '':
        return None
    return datetime.strptime(op_row.op_value, "%Y-%m-%d %H:%M:%S")


def record_last_sync_time(session, sync_time=None):
    if sync_time is None:
        sync_time_str = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    else:
        sync_time_str = sync_time.strftime("%Y-%m-%d %H:%M:%S")

    session.query(ib_models.InfobloxOperation).\
        filter_by(op_type='last_sync_time').\
        update({'op_value': sync_time_str})


# Neutron General Queries
def get_subnets_by_network_id(session, network_id):
    q = session.query(models_v2.Subnet).filter_by(network_id=network_id)
    return q.all()


def get_subnet_by_id(session, subnet_id):
    q = session.query(models_v2.Subnet).filter_by(id=subnet_id)
    return q.one()


def get_subnets_by_tenant_id(session, tenant_id):
    q = session.query(models_v2.Subnet).filter_by(tenant_id=tenant_id)
    return q.all()


def get_port_by_id(session, port_id):
    q = session.query(models_v2.Port)
    return q.filter_by(id=port_id).one()


def get_subnet_dhcp_port_address(session, subnet_id):
    dhcp_port = (session.query(models_v2.IPAllocation).
                 filter_by(subnet_id=subnet_id).
                 join(models_v2.Port).
                 filter_by(device_owner='network:dhcp')
                 .first())
    if dhcp_port:
        return dhcp_port.ip_address
    return None


def get_floatingip_by_id(session, floatingip_id):
    q = session.query(l3_db.FloatingIP)
    return q.filter_by(id=floatingip_id).one()


def get_floatingip_by_ip_address(session, floatingip):
    q = session.query(l3_db.FloatingIP)
    return q.filter_by(floating_ip_address=floatingip).one()


def get_address_scope_by_subnetpool_id(session, subnetpool_id):
    sub_qry = (session.query(models_v2.SubnetPool.address_scope_id).
               filter(models_v2.SubnetPool.id == subnetpool_id))
    q = (session.query(address_scope_db.AddressScope).
         filter(address_scope_db.AddressScope.id.in_(sub_qry)))
    return q.first()


def is_last_subnet(session, subnet_id):
    q = (session.query(models_v2.Subnet).
         filter(models_v2.Subnet.id != subnet_id))
    return q.count() == 0


def is_last_subnet_in_network(session, subnet_id, network_id):
    q = (session.query(models_v2.Subnet).
         filter(models_v2.Subnet.id != subnet_id,
                models_v2.Subnet.network_id == network_id))
    return q.count() == 0


def is_last_subnet_in_tenant(session, subnet_id, tenant_id):
    q = (session.query(models_v2.Subnet).
         filter(models_v2.Subnet.id != subnet_id,
                models_v2.Subnet.tenant_id == tenant_id))
    return q.count() == 0


def is_last_subnet_in_private_networks(session, subnet_id):
    sub_qry = session.query(external_net_db.ExternalNetwork.network_id)
    q = (session.query(models_v2.Subnet.id).
         filter(models_v2.Subnet.id != subnet_id).
         filter(~models_v2.Subnet.network_id.in_(sub_qry)))
    return q.count() == 0


def is_last_subnet_in_address_scope(session, subnet_id):
    q = (session.query(models_v2.Subnet).
         filter(models_v2.Subnet.subnetpool_id == models_v2.SubnetPool.id).
         filter(address_scope_db.id == models_v2.SubnetPool.address_scope_id).
         filter(models_v2.Subnet.id != subnet_id))
    return q.count() == 0


def add_tenant(session, tenant_id, tenant_name):
    tenant = ib_models.InfobloxTenant(
        tenant_id=tenant_id,
        tenant_name=tenant_name)
    session.add(tenant)
    return tenant


def add_or_update_tenant(session, tenant_id, tenant_name):
    db_tenant = get_tenant(session, tenant_id)
    if db_tenant is None:
        add_tenant(session, tenant_id, tenant_name)
    elif db_tenant.tenant_name != tenant_name:
        db_tenant.tenant_name = tenant_name


def get_tenant(session, tenant_id):
    q = session.query(ib_models.InfobloxTenant)
    return q.filter_by(tenant_id=tenant_id).first()


def get_tenants(session, tenant_ids=None):
    q = session.query(ib_models.InfobloxTenant)
    if tenant_ids:
        q = q.filter(ib_models.InfobloxTenant.tenant_id.in_(tenant_ids))
    return q.all()


def get_external_subnets(session):
    sub_qry = session.query(external_net_db.ExternalNetwork.network_id)
    return (session.query(models_v2.Subnet).
            filter(models_v2.Subnet.network_id.in_(sub_qry))).all()


def get_floatingip_ports(session, floating_ips, floating_network_id):
    q = (session.query(models_v2.Port.id,
                       models_v2.Port.device_id,
                       models_v2.Port.device_owner,
                       l3_db.FloatingIP.floating_ip_address,
                       models_v2.Port.name).
         filter(models_v2.Port.id == l3_db.FloatingIP.floating_port_id).
         filter(l3_db.FloatingIP.floating_ip_address.in_(floating_ips)))
    return q.all()


def add_instance(session, instance_id, instance_name):
    instance = ib_models.InfobloxInstance(
        instance_id=instance_id,
        instance_name=instance_name)
    session.add(instance)
    return instance


def add_or_update_instance(session, instance_id, instance_name):
    db_instance = get_instance(session, instance_id)
    if db_instance is None:
        add_instance(session, instance_id, instance_name)
    elif db_instance.instance_name != instance_name:
        db_instance.instance_name = instance_name


def get_instance(session, instance_id):
    q = session.query(ib_models.InfobloxInstance)
    return q.filter_by(instance_id=instance_id).first()


def remove_instance(session, instance_id):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxInstance)
        q = q.filter_by(instance_id=instance_id)
        q.delete(synchronize_session=False)


def get_instances(session, instance_ids=None):
    q = session.query(ib_models.InfobloxInstance)
    if instance_ids:
        q = q.filter(ib_models.InfobloxInstance.instance_id.in_(instance_ids))
    return q.all()


def add_network(session, network_id, network_name):
    network = ib_models.InfobloxNetwork(
        network_id=network_id,
        network_name=network_name)
    session.add(network)
    return network


def add_or_update_network(session, network_id, network_name):
    db_network = get_network(session, network_id)
    if db_network is None:
        add_network(session, network_id, network_name)
    elif db_network.network_name != network_name:
        db_network.network_name = network_name


def get_network(session, network_id):
    q = session.query(ib_models.InfobloxNetwork)
    return q.filter_by(network_id=network_id).first()


def remove_network(session, network_id):
    with session.begin(subtransactions=True):
        q = session.query(ib_models.InfobloxNetwork)
        q = q.filter_by(network_id=network_id)
        q.delete(synchronize_session=False)



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_config.types as types

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi


DELIMITER = '^'


class GridMappingManager(object):

    # network view mapping order listed hierarchically from bottom to top
    mapping_neutron_object_names = [const.EA_MAPPING_SUBNET_ID,
                                    const.EA_MAPPING_SUBNET_CIDR,
                                    const.EA_MAPPING_NETWORK_NAME,
                                    const.EA_MAPPING_NETWORK_ID,
                                    const.EA_MAPPING_TENANT_NAME,
                                    const.EA_MAPPING_TENANT_ID,
                                    const.EA_MAPPING_ADDRESS_SCOPE_NAME,
                                    const.EA_MAPPING_ADDRESS_SCOPE_ID]

    def __init__(self, grid_config):
        self._grid_config = grid_config
        self._connector = self._grid_config.gm_connector
        self._context = self._grid_config.context
        self._grid_id = self._grid_config.grid_id

    def sync(self):
        """Discovers and syncs networks between Neutron and Infoblox backend.

        The following information is discovered and synchronized.
        1. network views
        2. neutron network mapping conditions to NIOS network
        3. authority members that owns network views by either GM ownership or
           delegation to Cloud Platform Members (CPM).

        :return: None
        """
        session = self._context.session
        self.db_members = dbi.get_members(session, grid_id=self._grid_id)

        discovered_network_views = self._discover_network_views()
        if not discovered_network_views:
            return

        discovered_dns_views = self._discover_dns_views()
        dns_views = self.get_dns_views(discovered_dns_views)

        discovered_delegations = self._sync_network_views(
            discovered_network_views, dns_views)

        discovered_networks = self._discover_networks()
        self._sync_network_mapping(discovered_networks, discovered_delegations)

    def _load_persisted_mappings(self):
        session = self._context.session
        self.db_network_views = dbi.get_network_views(
            session, grid_id=self._grid_id)
        self.db_mapping_conditions = dbi.get_mapping_conditions(
            session, grid_id=self._grid_id)
        self.db_authority_members = dbi.get_mapping_members(
            session, grid_id=self._grid_id)
        self.db_service_members = dbi.get_service_members(
            session, grid_id=self._grid_id)

    def get_dns_views(self, discovered_dns_views):
        dns_views = dict()
        for dns_view in discovered_dns_views:
            netview_name = dns_view['network_view']
            dnsview_name = dns_view['name']
            if netview_name in dns_views and (
                    dns_views[netview_name] != self._grid_config.dns_view):
                dns_views[netview_name] = self._grid_config.dns_view
            if netview_name not in dns_views:
                dns_views[netview_name] = dnsview_name
        return dns_views

    def _sync_network_views(self, discovered_netviews, dns_views):
        """Discover network views and sync with db.

        The discovered network view json contains the following data:
        - network view
        - cloud_info for delegated member if cloud platform is supported
        - mapping conditional EAs

        So discovered information will be updated in tables such as
        infoblox_network_views and infoblox_mapping_conditions.

        :param discovered_netviews: discovered network view json
        :return: None
        """
        session = self._context.session
        self._load_persisted_mappings()
        discovered_delegations = dict()

        persisted_netview_ids = utils.get_values_from_records(
            'id', self.db_network_views)
        discovered_netview_ids = []

        for netview in discovered_netviews:
            netview_name = netview['name']
            is_default = netview[const.IS_DEFAULT]
            netview_id = utils.get_network_view_id(self._grid_id,
                                                   netview['_ref'])

            cloud_adapter_id_vals = utils.get_ea_value(
                const.EA_CLOUD_ADAPTER_ID, netview, True)
            if cloud_adapter_id_vals is None:
                participated = False
            else:
                cloud_adapter_ids = [gid for gid in cloud_adapter_id_vals
                                     if int(gid) == self._grid_id]
                participated = True if cloud_adapter_ids else False

            if not participated:
                continue

            shared_val = utils.get_ea_value(const.EA_IS_SHARED, netview)
            is_shared = types.Boolean()(shared_val) if shared_val else False

            # authority member is default to GM
            gm_row = utils.find_one_in_list('member_type',
                                            const.MEMBER_TYPE_GRID_MASTER,
                                            self.db_members)
            authority_member_id = gm_row.member_id

            # get delegation member if cloud platform is supported
            delegated_member = self._get_delegated_member(netview)
            if delegated_member:
                authority_member_id = delegated_member.member_id
                discovered_delegations[netview_name] = (
                    delegated_member.member_id)

            dns_view = (dns_views[netview_name] if dns_views.get(netview_name)
                        else None)

            # see if the network view already exists in db
            netview_row = utils.find_one_in_list('id',
                                                 netview_id,
                                                 self.db_network_views)
            if netview_row:
                dbi.update_network_view(session, netview_id, netview_name,
                                        authority_member_id, is_shared,
                                        dns_view, participated, is_default)
            else:
                internal_netview = (const.DEFAULT_NETWORK_VIEW if is_default
                                    else netview_name)
                internal_dnsview = (const.DEFAULT_DNS_VIEW if is_default
                                    else dns_view)
                dbi.add_network_view(session,
                                     netview_id,
                                     netview_name,
                                     self._grid_id,
                                     authority_member_id,
                                     is_shared,
                                     dns_view,
                                     internal_netview,
                                     internal_dnsview,
                                     participated,
                                     is_default)

            discovered_netview_ids.append(netview_id)

            # update mapping conditions for the current network view
            self._update_mapping_conditions(netview, netview_id, participated)

        # we have added new network views. now let's remove persisted
        # network views not found from discovery
        persisted_set = set(persisted_netview_ids)
        removable_set = persisted_set.difference(discovered_netview_ids)
        removable_netviews = list(removable_set)
        if removable_netviews:
            dbi.remove_network_views(session, removable_netviews)
        session.flush()
        return discovered_delegations

    def _sync_network_mapping(self, discovered_networks,
                              discovered_delegations):
        """Discover networks and sync with db.

        The discovered network json contains the following data:
        - network view
        - network
        - cloud_info for delegated member if cloud platform is supported
        - dhcp members

        :param discovered_networks: discovered network json
        :param discovered_delegations: discovered delegation members
        :return: None
        """
        session = self._context.session
        self._load_persisted_mappings()

        discovered_mapping = self._get_member_mapping(discovered_networks,
                                                      discovered_delegations)

        # add or remove authority mapping members
        persisted_authority_members = utils.get_composite_values_from_records(
            ['network_view_id', 'member_id', 'mapping_relation'],
            self.db_authority_members,
            DELIMITER)
        persisted_set = set(persisted_authority_members)
        discovered_set = set(discovered_mapping['authority_members'])
        addable_set = discovered_set.difference(persisted_set)
        removable_set = persisted_set.difference(discovered_set)

        for authority_member_info in addable_set:
            authority_member = authority_member_info.split(DELIMITER)
            network_view_id = authority_member[0]
            member_id = authority_member[1]
            mapping_relation = authority_member[2]
            dbi.add_mapping_member(session, network_view_id, member_id,
                                   mapping_relation)

        for authority_member_info in removable_set:
            authority_member = authority_member_info.split(DELIMITER)
            network_view_id = authority_member[0]
            member_id = authority_member[1]
            dbi.remove_mapping_member(session, network_view_id, member_id)

        # add or remove service members
        persisted_service_members = utils.get_composite_values_from_records(
            ['network_view_id', 'member_id', 'service'],
            self.db_service_members,
            DELIMITER)
        persisted_set = set(persisted_service_members)
        discovered_set = set(discovered_mapping['service_members'])
        addable_set = discovered_set.difference(persisted_set)
        removable_set = persisted_set.difference(discovered_set)

        for service_member_info in addable_set:
            service_member = service_member_info.split(DELIMITER)
            network_view_id = service_member[0]
            member_id = service_member[1]
            service = service_member[2]
            dbi.add_service_member(session, network_view_id, member_id,
                                   service)

        for service_member_info in removable_set:
            service_member = service_member_info.split(DELIMITER)
            network_view_id = service_member[0]
            member_id = service_member[1]
            service = service_member[2]
            dbi.remove_service_member(session, network_view_id,
                                      member_id=member_id, service=service)

    def _discover_network_views(self):
        return_fields = ['name', 'is_default', 'extattrs']
        if self._grid_config.is_cloud_wapi:
            return_fields.append('cloud_info')

        netviews = self._connector.get_object('networkview',
                                              return_fields=return_fields)
        if not netviews:
            return []
        return netviews

    def _discover_networks(self):
        return_fields = ['members', 'network_view', 'network', 'options']
        if self._grid_config.is_cloud_wapi:
            return_fields.append('cloud_info')

        # TODO(pbondar): Consider using NetworkV4 and NetworkV6 objects
        #                from infoblox-client to interact with NIOS
        ipv4networks = self._connector.get_object('network',
                                                  return_fields=return_fields)
        ipv6networks = self._connector.get_object('ipv6network',
                                                  return_fields=return_fields)
        # get_object returns None if nothing was found, so convert results
        if not ipv4networks:
            ipv4networks = []
        if not ipv6networks:
            ipv6networks = []
        return ipv4networks + ipv6networks

    def _discover_dns_views(self):
        return_fields = ['name', 'network_view']
        dns_views = self._connector.get_object('view',
                                               return_fields=return_fields)
        if not dns_views:
            return []
        return dns_views

    def _get_member_mapping(self, discovered_networks, discovered_delegations):
        """Returns members that are used for authority and dhcp.

        Authority members own network views because they are either GM who owns
        non delegated network views or Cloud Platform Member(CPM) who owns
        delegated network views.

        DHCP members are the members who serve DHCP protocols.
        """
        gm_row = utils.find_one_in_list('member_type',
                                        const.MEMBER_TYPE_GRID_MASTER,
                                        self.db_members)
        gm_member_id = gm_row.member_id
        mapping_authority_members = []
        mapping_service_members = []

        # first get delegated authority members from Infoblox network views
        for netview in discovered_delegations:
            netview_row = utils.find_one_in_list('network_view', netview,
                                                 self.db_network_views)
            if not netview_row:
                continue
            netview_id = netview_row.id
            authority_member = discovered_delegations[netview]
            mapping_relation = const.MAPPING_RELATION_DELEGATED
            mapping_row_info = (netview_id + DELIMITER +
                                authority_member + DELIMITER +
                                mapping_relation)
            mapping_authority_members.append(mapping_row_info)

        # then get authority and dhcp members from Infoblox networks
        for network in discovered_networks:
            netview = network['network_view']
            netview_row = utils.find_one_in_list('network_view', netview,
                                                 self.db_network_views)
            if not netview_row or not netview_row.participated:
                continue

            netview_id = netview_row.id

            # get authority member
            mapping_relation = const.MAPPING_RELATION_GM_OWNED
            authority_member = gm_member_id
            delegated_member = self._get_delegated_member(network)

            if netview in discovered_delegations:
                authority_member = discovered_delegations[netview]
                mapping_relation = const.MAPPING_RELATION_DELEGATED
            elif delegated_member:
                authority_member = delegated_member.member_id
                mapping_relation = const.MAPPING_RELATION_DELEGATED

            mapping_member_info = (netview_id + DELIMITER +
                                   authority_member + DELIMITER +
                                   mapping_relation)
            if mapping_member_info not in mapping_authority_members:
                mapping_authority_members.append(mapping_member_info)

            # get dhcp member
            dhcp_members = self._get_dhcp_members(network)
            for member in dhcp_members:
                mapping_member_info = (netview_id + DELIMITER +
                                       member.member_id + DELIMITER +
                                       const.SERVICE_TYPE_DHCP)
                if mapping_member_info not in mapping_service_members:
                    mapping_service_members.append(mapping_member_info)

            # get dns member
            dns_members = self._get_dns_members(network)
            for member in dns_members:
                mapping_member_info = (netview_id + DELIMITER +
                                       member.member_id + DELIMITER +
                                       const.SERVICE_TYPE_DNS)
                if mapping_member_info not in mapping_service_members:
                    mapping_service_members.append(mapping_member_info)

        return {'authority_members': mapping_authority_members,
                'service_members': mapping_service_members}

    def _get_mapping_conditions(self, netview_dict):
        conditions = dict()
        for object_name in self.mapping_neutron_object_names:
            object_value = utils.get_ea_value(object_name, netview_dict, True)
            if object_value:
                conditions[object_name] = object_value
        return conditions

    def _get_delegated_member(self, network_dict):
        delegated_member = None
        if (network_dict.get('cloud_info') and
                network_dict['cloud_info'].get('delegated_member')):
            delegated_member_name = (
                network_dict['cloud_info']['delegated_member']['name'])
            delegated_member = utils.find_one_in_list(
                'member_name', delegated_member_name, self.db_members)
            if not delegated_member:
                raise exc.InfobloxCannotFindMember(
                    member=delegated_member_name)
        return delegated_member

    def _get_dhcp_members(self, network_dict):
        # multiple dhcp members can be assigned to a network
        dhcp_members = []
        member_ips = utils.get_dhcp_member_ips(network_dict)
        for member_ip in member_ips:
            dhcp_member = utils.find_in_list_by_value(member_ip,
                                                      self.db_members)
            if not dhcp_member:
                raise exc.InfobloxCannotFindMember(
                    member=member_ip)
            dhcp_members.append(dhcp_member)
        return dhcp_members

    def _get_dns_members(self, network_dict):
        # multiple dns members can be assigned to a network
        dns_members = []
        member_ips = utils.get_dns_member_ips(network_dict)
        for member_ip in member_ips:
            dns_member = utils.find_in_list_by_value(member_ip,
                                                     self.db_members)
            if dns_member:
                dns_members.append(dns_member)
        return dns_members

    def _update_mapping_conditions(self, discovered_netview, netview_id,
                                   participated):
        session = self._context.session
        mapping_conditions = dict()
        if participated:
            mapping_conditions = self._get_mapping_conditions(
                discovered_netview)
        discovered_condition_rows = []
        for condition_name in mapping_conditions:
            conditions = [netview_id + DELIMITER + condition_name + DELIMITER +
                          value for value in
                          mapping_conditions[condition_name]]
            discovered_condition_rows += conditions

        mapping_condition_rows = utils.find_in_list(
            'network_view_id', [netview_id], self.db_mapping_conditions)
        condition_rows = utils.get_composite_values_from_records(
            ['network_view_id', 'neutron_object_name', 'neutron_object_value'],
            mapping_condition_rows,
            DELIMITER)
        persisted_set = set(condition_rows)
        discovered_set = set(discovered_condition_rows)
        addable_set = discovered_set.difference(persisted_set)
        removable_set = persisted_set.difference(discovered_set)

        for condition_attr in addable_set:
            condition = condition_attr.split(DELIMITER)
            network_view_id = condition[0]
            neutron_object_name = condition[1]
            neutron_object_value = condition[2]
            dbi.add_mapping_condition(session,
                                      network_view_id,
                                      neutron_object_name,
                                      neutron_object_value)

        for condition_attr in removable_set:
            condition = condition_attr.split(DELIMITER)
            network_view_id = condition[0]
            neutron_object_name = condition[1]
            neutron_object_value = condition[2]
            dbi.remove_mapping_condition(session,
                                         network_view_id,
                                         neutron_object_name,
                                         neutron_object_value)



# Copyright 2016 Infoblox Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""store_instances_names_in_db

Revision ID: 5267e17f370f
Revises: 4d0bb1d080f8
Create Date: 2016-06-17 20:55:22.736519

"""

# revision identifiers, used by Alembic.
revision = '5267e17f370f'
down_revision = '4d0bb1d080f8'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'infoblox_instances',
        sa.Column('instance_id', sa.String(64),
                  nullable=False, primary_key=True),
        sa.Column('instance_name', sa.String(255), nullable=False),
    )



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron_lib import constants as n_const

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import pattern

from networking_infoblox.tests import base


class TestPatternBuilder(base.TestCase):

    def setUp(self):
        super(TestPatternBuilder, self).setUp()
        self.ib_cxt = self._get_ib_context()
        self.pattern_builder = pattern.PatternBuilder(self.ib_cxt)
        self.test_ip = '11.11.11.11'
        self.expected_ip = self.test_ip.replace('.', '-').replace(':', '-')
        self.expected_domain = (
            self.ib_cxt.grid_config.default_domain_name_pattern.replace(
                '{subnet_id}', self.ib_cxt.subnet['id'])
        )

    def _get_ib_context(self):
        ib_cxt = mock.Mock()
        ib_cxt.network = {'id': 'network-id',
                          'name': 'test-net-1',
                          'tenant_id': 'network-id'}
        ib_cxt.subnet = {'id': 'subnet-id',
                         'name': 'test-sub-1',
                         'tenant_id': 'tenant-id',
                         'network_id': 'network-id'}
        ib_cxt.grid_config.default_host_name_pattern = 'host-{ip_address}'
        ib_cxt.grid_config.default_domain_name_pattern = (
            '{subnet_id}.infoblox.com')
        ib_cxt.grid_config.external_host_name_pattern = ''
        ib_cxt.grid_config.external_domain_name_pattern = ''
        return ib_cxt

    def _get_test_port(self, device_owner):
        return {'id': 'port-id',
                'name': 'port-name',
                'device_owner': device_owner,
                'port_id': 'port-id',
                'device_id': 'device-id'}

    def _test_get_hostname(self, expected_hostname,
                           device_owner=n_const.DEVICE_OWNER_FLOATINGIP,
                           instance_name='test-vm',
                           external=False,):
        test_port = self._get_test_port(device_owner)
        actual_hostname = self.pattern_builder.get_hostname(
            self.test_ip, instance_name, test_port['id'],
            test_port['device_owner'], test_port['device_id'],
            external=external)
        self.assertEqual(expected_hostname, actual_hostname)

    def test_get_hostname_for_floating_ip_with_instance_name(self):
        expected_hostname = str.format("floating-ip-{}.{}", self.expected_ip,
                                       self.expected_domain)
        self._test_get_hostname(expected_hostname, external=True)

    def test_get_hostname_for_floating_ip_without_instance_name(self):
        instance_name = None
        expected_hostname = str.format("floating-ip-{}.{}", self.expected_ip,
                                       self.expected_domain)
        self._test_get_hostname(expected_hostname,
                                instance_name=instance_name,
                                external=True)

    def test_get_hostname_for_floating_ip_with_instance_name_pattern(self):
        instance_name = 'test-vm'
        self.pattern_builder.grid_config.default_host_name_pattern = (
            'host-{instance_name}')
        expected_hostname = str.format("host-{}.{}", instance_name,
                                       self.expected_domain)
        self._test_get_hostname(expected_hostname,
                                instance_name=instance_name,
                                external=True)

    def test_get_hostname_for_floating_ip_external_pattern(self):
        self.pattern_builder.grid_config.external_host_name_pattern = (
            '{instance_name}')
        self.pattern_builder.grid_config.external_domain_name_pattern = (
            'external.infoblox.com')
        instance_name = 'test-vm'
        expected_hostname = '.'.join([
            instance_name,
            self.pattern_builder.grid_config.external_domain_name_pattern])
        self._test_get_hostname(expected_hostname,
                                instance_name=instance_name,
                                external=True)

    def test_get_hostname_for_other_device_owners(self):
        for dev, patt in const.NEUTRON_DEVICE_OWNER_TO_PATTERN_MAP.items():
            expected_hostname = str.format(
                "{}.{}", patt.replace('{ip_address}',
                                      self.expected_ip),
                self.expected_domain)
            self._test_get_hostname(expected_hostname,
                                    device_owner=dev,
                                    instance_name='')

    def test_get_hostname_for_other_device_owners_external(self):
        self.pattern_builder.grid_config.external_host_name_pattern = (
            '{instance_name}')
        self.pattern_builder.grid_config.external_domain_name_pattern = (
            'external.infoblox.com')
        for dev, patt in const.NEUTRON_DEVICE_OWNER_TO_PATTERN_MAP.items():
            expected_hostname = str.format(
                "{}.{}", patt.replace('{ip_address}',
                                      self.expected_ip),
                self.pattern_builder.grid_config.external_domain_name_pattern)
            self._test_get_hostname(expected_hostname,
                                    device_owner=dev,
                                    instance_name='',
                                    external=True)

    def test_get_hostname_for_instance_name(self):
        test_port = self._get_test_port('')
        self.pattern_builder.grid_config.default_host_name_pattern = (
            'host-{instance_name}')

        instance_name = 'test.vm'
        actual_hostname = self.pattern_builder.get_hostname(
            self.test_ip, instance_name, test_port['id'],
            test_port['device_owner'], test_port['device_id'])
        expected_instance_name = (
            instance_name.replace('.', '-').replace(':', '-'))
        expected_hostname = str.format("host-{}.{}", expected_instance_name,
                                       self.expected_domain)
        self.assertEqual(expected_hostname, actual_hostname)

    def test_get_hostname_for_port_name(self):
        test_port = self._get_test_port('')
        self.pattern_builder.grid_config.default_host_name_pattern = (
            'host-{port_name}')

        # test hostname with port_name
        actual_hostname = self.pattern_builder.get_hostname(
            self.test_ip, None, test_port['id'], test_port['device_owner'],
            test_port['device_id'], test_port['name'])
        expected_hostname = str.format("host-{}.{}", test_port['name'],
                                       self.expected_domain)
        self.assertEqual(expected_hostname, actual_hostname)

        # test hostname without port_name but with ip
        actual_hostname = self.pattern_builder.get_hostname(
            self.test_ip, None, test_port['id'], test_port['device_owner'],
            test_port['device_id'])
        ip_addr = self.test_ip.replace('.', '-').replace(':', '-')
        expected_hostname = str.format("host-{}.{}", ip_addr,
                                       self.expected_domain)
        self.assertEqual(expected_hostname, actual_hostname)

        # test hostname without port_name and ip
        actual_hostname = self.pattern_builder.get_hostname(
            None, None, test_port['id'], test_port['device_owner'],
            test_port['device_id'])
        expected_hostname = str.format("host-{}.{}", test_port['id'],
                                       self.expected_domain)
        self.assertEqual(expected_hostname, actual_hostname)

    def test_get_zone_name(self):
        # test {subnet_id} pattern
        actual_zone = self.pattern_builder.get_zone_name()
        self.assertEqual(self.expected_domain, actual_zone)

        # test {network_name} pattern
        self.pattern_builder.grid_config.default_domain_name_pattern = (
            '{network_name}.infoblox.com')
        self.expected_domain = (
            self.ib_cxt.grid_config.default_domain_name_pattern.replace(
                '{network_name}', self.ib_cxt.network['name'])
        )
        actual_zone = self.pattern_builder.get_zone_name()
        self.assertEqual(self.expected_domain, actual_zone)

        # test static zone name
        zone = 'infoblox.com'
        self.pattern_builder.grid_config.default_domain_name_pattern = zone
        actual_zone = self.pattern_builder.get_zone_name()
        self.assertEqual(zone, actual_zone)

    def test_get_zone_name_for_external_patern_not_set(self):
        actual_zone = self.pattern_builder.get_zone_name(is_external=True)
        # default pattern should be used if external is not set
        self.assertEqual(self.expected_domain, actual_zone)

    def test_get_zone_name_for_external_patern_set(self):
        external_pattern = 'external.infoblox.com'
        self.pattern_builder.grid_config.external_domain_name_pattern = (
            external_pattern)
        actual_zone = self.pattern_builder.get_zone_name(is_external=True)
        # external pattern should be used
        self.assertEqual(external_pattern, actual_zone)

    def test_get_zone_name_pattern(self):
        # test get_zone_name pattern
        self.assertEqual(
            self.pattern_builder.get_zone_name_pattern(is_external=False),
            self.ib_cxt.grid_config.default_domain_name_pattern)

    def test_get_zone_name_pattern_for_external(self):
        # test get_zone_name pattern for external zone
        self.pattern_builder.grid_config.external_domain_name_pattern = (
            'external.infoblox.com')
        self.assertEqual(
            self.pattern_builder.get_zone_name_pattern(is_external=True),
            self.ib_cxt.grid_config.external_domain_name_pattern)

    def test_validate_pattern_struct(self):
        pattern_dict = {
            'network_id': 'subnet',
            'network_name': 'network',
            'tenant_id': 'tenant_id',
            'tenant_name': 'tenant_name',
            'subnet_name': 'subnet_name',
            'subnet_id': 'subnet_id'
        }
        pattern = '{tenant_id}.{tenant_name}.{subnet_name}'
        self.pattern_builder._validate_pattern_struct(pattern, pattern_dict)
        self.assertTrue(True)

    def test_validate_incorrect_pattern_struct(self):
        pattern_dict = {
            'network_id': 'subnet',
            'network_name': 'network',
            'tenant_id': 'tenant_id',
            'tenant_name': 'tenant_name',
            'subnet_name': 'subnet_name',
            'subnet_id': 'subnet_id'
        }
        pattern = '{tenant_id}.{tenant_name}.{subnet_name_invalid}'
        self.assertRaises(KeyError,
                          self.pattern_builder._validate_pattern_struct,
                          pattern,
                          pattern_dict)



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_config.types as types
from oslo_serialization import jsonutils

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi


class GridMemberManager(object):

    def __init__(self, grid_config):
        self._grid_config = grid_config
        self._context = self._grid_config.context
        self._connector = self._grid_config.gm_connector

    def sync(self):
        """Discover and sync the active grid and its members."""
        self.sync_grid()
        self.sync_members()

    def sync_grid(self):
        """Synchronize an active grid.

        Only one active grid should be kept where grid_status is set to 'ON'.
        """
        session = self._context.session
        grid_connection = self._grid_config.get_grid_connection()
        grid_connection_json = jsonutils.dumps(grid_connection)

        db_grids = dbi.get_grids(session)
        db_grid_ids = utils.get_values_from_records('grid_id', db_grids)

        # update the existing grid or add new grid
        if self._grid_config.grid_id in db_grid_ids:
            dbi.update_grid(session,
                            self._grid_config.grid_id,
                            self._grid_config.grid_name,
                            grid_connection_json,
                            const.GRID_STATUS_ON)
        else:
            dbi.add_grid(session,
                         self._grid_config.grid_id,
                         self._grid_config.grid_name,
                         grid_connection_json,
                         const.GRID_STATUS_ON,
                         utils.get_hash())

        # deleting grids are delicate operation so we won't allow it
        # but we will set grid status to OFF to unused grids.
        persisted_set = set(db_grid_ids)
        disable_set = persisted_set.difference([self._grid_config.grid_id])
        disabling_grid_ids = list(disable_set)
        for grid_id in disabling_grid_ids:
            dbi.update_grid(session,
                            grid_id,
                            grid_status=const.GRID_STATUS_OFF)
        session.flush()

    def sync_members(self):
        """Synchronizes grid members.

        Members in the active grid are discovered from NIOS backend and
        grid members are in sync in neutron db. The members who are no longer
        in used are set to 'OFF' status.
        """
        session = self._context.session
        grid_id = self._grid_config.grid_id

        db_grids = dbi.get_grids(session)
        db_grid = utils.find_one_in_list('grid_id', grid_id, db_grids)
        gm_member_id = db_grid.gm_id

        db_members = dbi.get_members(session, grid_id=grid_id)
        gm_member = utils.find_one_in_list('member_id', gm_member_id,
                                           db_members)

        discovered_members = self._discover_members()
        if not discovered_members:
            return

        dns_member_settings = self._discover_dns_settings()
        dhcp_member_settings = self._discover_dhcp_settings()

        discovered_licenses = self._discover_member_licenses()

        discovered_member_ids = []

        for member in discovered_members:
            member_name = member['host_name']
            member_ip, member_ipv6 = self._get_lan1_ips(member)
            member_wapi = member_ip if member_ip else member_ipv6
            member_hwid = member['node_info'][0].get('hwid')
            member_status = self._get_member_status(
                member['node_info'][0]['service_status'])
            member_type = self._get_member_type(discovered_licenses,
                                                member_name,
                                                member_hwid)

            require_db_update = False
            if member_type == const.MEMBER_TYPE_GRID_MASTER:
                if gm_member:
                    require_db_update = True
                member_id = gm_member_id
                member_wapi = self._grid_config.grid_master_host
            else:
                # no need to process 'Is Cloud Member' flag for non GM members
                ea_is_cloud_member = utils.get_ea_value(
                    const.EA_IS_CLOUD_MEMBER, member)
                is_cloud_member = (types.Boolean()(ea_is_cloud_member)
                                   if ea_is_cloud_member else False)
                if not is_cloud_member:
                    continue

                db_member = utils.find_one_in_list('member_name', member_name,
                                                   db_members)
                if db_member:
                    require_db_update = True
                    member_id = db_member.member_id
                else:
                    member_id = utils.get_hash(str(grid_id) + member_name)

            member_dhcp_ip, member_dhcp_ipv6 = self._get_dhcp_ips(
                member, dhcp_member_settings)
            member_dns_ip, member_dns_ipv6 = self._get_dns_ips(
                member, dns_member_settings)

            if require_db_update:
                dbi.update_member(session,
                                  member_id,
                                  grid_id,
                                  member_name,
                                  member_ip,
                                  member_ipv6,
                                  member_type,
                                  member_status,
                                  member_dhcp_ip,
                                  member_dhcp_ipv6,
                                  member_dns_ip,
                                  member_dns_ipv6,
                                  member_wapi)
            else:
                dbi.add_member(session,
                               member_id,
                               grid_id,
                               member_name,
                               member_ip,
                               member_ipv6,
                               member_type,
                               member_status,
                               member_dhcp_ip,
                               member_dhcp_ipv6,
                               member_dns_ip,
                               member_dns_ipv6,
                               member_wapi)

            discovered_member_ids.append(member_id)

        # deleting members are delicate operation so we won't allow it
        # but we will set member status to OFF to unused members.
        db_member_ids = utils.get_values_from_records('member_id', db_members)
        persisted_set = set(db_member_ids)
        discovered_set = set(discovered_member_ids)
        disable_set = persisted_set.difference(discovered_set)
        disabling_member_ids = list(disable_set)
        for member_id in disabling_member_ids:
            dbi.update_member(session,
                              member_id,
                              grid_id,
                              member_status=const.MEMBER_STATUS_OFF)
        session.flush()

    def _discover_members(self):
        return_fields = ['node_info', 'host_name', 'vip_setting', 'extattrs']
        # ipv6_setting, lan2_port_setting and mgmt_port_setting fields are
        # available with wapi 2.2+, so check only 'member_ipv6_setting' feature
        if utils.get_features(
                self._grid_config.wapi_version).member_ipv6_setting:
            return_fields.extend(['ipv6_setting', 'lan2_port_setting',
                                  'mgmt_port_setting'])

        members = self._connector.get_object('member',
                                             return_fields=return_fields)
        return members

    def _discover_dns_settings(self):
        members = {}
        # all this info available only with wapi 2.3+
        features = utils.get_features(self._grid_config.wapi_version)
        if not features.dns_settings:
            return members

        return_fields = ['host_name', 'use_mgmt_port', 'use_mgmt_ipv6_port',
                         'use_lan_port', 'use_lan_ipv6_port', 'use_lan2_port',
                         'use_lan2_ipv6_port', 'additional_ip_list']
        dns_members = self._connector.get_object('member:dns',
                                                 return_fields=return_fields)
        # Convert members into dict with host_name as a key
        if dns_members:
            return {member['host_name']: member for member in dns_members}
        return members

    def _discover_dhcp_settings(self):
        members = {}
        # enable_dhcp available only with wapi 2.2.1+
        features = utils.get_features(self._grid_config.wapi_version)
        if not features.enable_dhcp:
            return members

        return_fields = ['host_name', 'enable_dhcp']
        dhcp_members = self._connector.get_object('member:dhcpproperties',
                                                  return_fields=return_fields)
        # Convert members into dict with host_name as a key
        if dhcp_members:
            return {member['host_name']: member for member in dhcp_members}
        return members

    def _get_lan1_ips(self, member):
        member_ip = member['vip_setting']['address']
        member_ipv6 = (member['ipv6_setting'].get('virtual_ip')
                       if member.get('ipv6_setting') else None)
        return member_ip, member_ipv6

    def _get_dhcp_ips(self, member, dhcp_member_settings):
        member_dhcp_ip = None
        member_dhcp_ipv6 = None
        member_name = member['host_name']
        member_ip, member_ipv6 = self._get_lan1_ips(member)

        if member_name in dhcp_member_settings:
            # Use LAN1 interface if enable_dhcp is turned on or LAN2 is
            # not configured
            l2 = member.get('lan2_port_setting')
            if dhcp_member_settings[member_name].get('enable_dhcp') or not l2:
                member_dhcp_ip = member_ip
                member_dhcp_ipv6 = member_ipv6
            else:
                if l2.get('network_setting'):
                    member_dhcp_ip = l2['network_setting'].get('address')
                if l2.get('v6_network_setting'):
                    member_dhcp_ipv6 = l2['v6_network_setting'].get(
                        'virtual_ip')
        return member_dhcp_ip, member_dhcp_ipv6

    def _get_dns_ips(self, member, dns_member_settings):
        member_dns_ip = None
        member_dns_ipv6 = None
        member_name = member['host_name']
        member_ip, member_ipv6 = self._get_lan1_ips(member)

        if member_name in dns_member_settings:
            dns_settings = dns_member_settings[member_name]
            # Assign IPv4 address
            if dns_settings.get('use_lan_port'):
                member_dns_ip = member_ip
            elif dns_settings.get('use_lan2_port'):
                l2 = member.get('lan2_port_setting')
                if l2 and l2.get('network_setting'):
                    member_dns_ip = l2['network_setting'].get('address')
            elif dns_settings.get('use_mgmt_port'):
                n_info = member.get('node_info')
                if n_info and n_info[0].get('mgmt_network_setting'):
                    member_dns_ip = n_info[0]['mgmt_network_setting'].get(
                        'address')
            elif dns_settings.get('additional_ip_list'):
                for ip in dns_settings.get('additional_ip_list'):
                    if utils.get_ip_version(ip) == 4:
                        member_dns_ip = ip
                        break

            # If ip is still blank fallback to member ip
            if not member_dns_ip:
                member_dns_ip = member_ip

            # Assign IPv6 address
            if dns_settings.get('use_lan_ipv6_port'):
                member_dns_ipv6 = member_ipv6
            elif dns_settings.get('use_lan2_ipv6_port'):
                l2 = member.get('lan2_port_setting')
                if l2 and l2.get('v6_network_setting'):
                    member_dns_ipv6 = l2['v6_network_setting'].get(
                        'virtual_ip')
            elif dns_settings.get('use_mgmt_ipv6_port'):
                n_info = member.get('node_info')
                if n_info and n_info[0].get('v6_mgmt_network_setting'):
                    member_dns_ipv6 = n_info[0]['v6_mgmt_network_setting'].get(
                        'virtual_ip')
            elif dns_settings.get('additional_ip_list'):
                for ip in dns_settings.get('additional_ip_list'):
                    if utils.get_ip_version(ip) == 6:
                        member_dns_ipv6 = ip
                        break

            # If ipv6 is still blank fallback to member ipv6
            if not member_dns_ipv6:
                member_dns_ipv6 = member_ipv6

        return member_dns_ip, member_dns_ipv6

    def _discover_member_licenses(self):
        if not utils.get_features(
                self._grid_config.wapi_version).member_licenses:
            return None

        return_fields = ['expiry_date', 'hwid', 'kind', 'type']
        licenses = self._connector.get_object('member:license',
                                              return_fields=return_fields)
        return licenses

    def _get_gm_info(self):
        """Get detail GM info.

        'grid_master_host' configuration accepts host IP or name of GM, so
        we need to figure whether hostname is used or ip address for either
        ipv4 or ipv6.
        """
        gm_ipv4 = None
        gm_ipv6 = None
        gm_hostname = None

        gm_host = self._grid_config.grid_master_host
        if utils.is_valid_ip(gm_host):
            ip_version = utils.get_ip_version(gm_host)
            if ip_version == 4:
                gm_ipv4 = gm_host
            else:
                gm_ipv6 = gm_host
        else:
            gm_hostname = gm_host

        return {'ipv4': gm_ipv4, 'ipv6': gm_ipv6, 'host': gm_hostname}

    def _get_member_status(self, member_service_status):
        node_status = None
        for ns in member_service_status:
            if ns['service'] == 'NODE_STATUS':
                node_status = ns['status']
                break
        return utils.get_member_status(node_status)

    def _get_member_type(self, member_licenses, member_name, member_hwid):
        if self._grid_config.grid_master_name == member_name:
            return const.MEMBER_TYPE_GRID_MASTER

        # member is not GM, so figure out whether the member is CPM or REGULAR
        # for cloud, 'CLOUD_API' license must exist to qualify for CPM
        member_type = const.MEMBER_TYPE_REGULAR_MEMBER
        found_cloud_license = False
        if self._grid_config.is_cloud_wapi and member_licenses:
            for ml in member_licenses:
                if (ml['hwid'] == member_hwid and
                        ml['type'] == const.MEMBER_LICENSE_TYPE_CLOUD_API):
                    found_cloud_license = True
                    break
            if found_cloud_license:
                member_type = const.MEMBER_TYPE_CP_MEMBER

        return member_type
